#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from six import string_types
from xarray import Dataset


class DataRecord(object):
    """Data structure to store variables in time and/or space dimensions.

    This class uses a xarray Dataset to store variables. This datastructure is
    located at the property ``dataset``. The DataRecord improves on an xarray
    Dataset by providing additional attributes and functions, including the
    ability to aggregate values on Landlab grid elements.

    Data variables can vary along one or both of the following dimensions:
        - time (model time)
        - item_id: variables can characterize a set of items (each identified
            by an individual id) that reside on the grid.

    Examples:
        - the variable 'mean_elevation' characterizes the grid and varies with
            time,
        - the variable 'clast__rock_type' characterizes a set of items (clasts)
            and varies with item_id,
        - the variable 'clast__size' can vary with both time and item_id

    If an item or set of items is defined, each item must be defined by the
    grid element and the element id at which it resides, e.g.:

        grid_element = 'node'
        element_id = 9.

    In the above case, `grid_element` and `element_id` are default data
    variables (in addition to any user-specified variables).

    For each item, `element_id` must be less than the number of this item's
    grid_element that exist on the grid. For example, if the grid has 100
    links, no item can live at link 100 or link -3 because only links 0 to 99
    exist in this example.
    """

    _name = "DataRecord"

    def __init__(self, grid, time=None, items=None, data_vars=None, attrs=None):
        """
        Parameters
        ----------
        grid : ModelGrid
        time : list or 1-D array of float or int (optional)
            The initial time(s) to add to the record. A time dimension is not
            created if the value is 'None' (default).
        items : dict (optional)
            Generic items that live on grid elements. No item is created if the
            value is 'None' (default). Otherwise, dictionary describes the
            position of generic items on the grid. The structure is:
                {'grid_element' : [grid_element],
                 'element_id' : [element_id]}
            where:
                - [grid_element] is a str or number-of-items-long array
                containing strings of the grid element(s) on which the items
                live. Valid locations depend on the grid type. If provided as a
                string it is assumed that all items live on the same type of
                grid element.
                - [element_id] is an array of integers identifying the grid
                element ID on which each item resides.
            An example argument would be:
                {'grid_element' : numpy.array(['node'], ['node'], ['link']),
                 'element_id' :   numpy.array([1],      [5],      [1]     )}
        data_vars : dict (optional)
            Dictionary of the data variables to be recorded. The structure is:
                {'variable_name_1' : (['dimensions'], variable_data_1),
                 'variable_name_2' : (['dimensions'], variable_data_2)}
            where:
                - 'variable_name...' is a string of the variable name (label)
                - ['dimensions'] is the dimension(s) over which the variable
                exists: can be ['time'], ['item_id'] or ['item_id', 'time'].
                - variable_data is an array containing the data, its size must
                match that of the variable dimension(s).
        attrs : dict (optional)
            Dictionary of global attributes on the DataRecord (metadata).
            Example: {'time_units' : 'y'}

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with time as the only dimension:
        >>> dr1 = DataRecord(grid, time=[0.],
        ...                  data_vars={'mean_elevation' : (['time'],
        ...                                                 np.array([100]))},
        ...                  attrs={'time_units' : 'y'})

        DataRecord is a xarray.Dataset, a multi-dimensional, in memory, array
        database. Dataset implements the mapping interface with keys given by
        variable names and values given by DataArray objects for each variable
        name.
        A DataRecord can have dimensions 'time' and/or 'item_id'.
        Coordinates are one dimensional arrays used for label-based indexing.
        DataRecord inherits all the methods and attributes from xarray.Dataset.

        >>> dr1.dataset.to_dataframe()
              mean_elevation
        time
        0.0              100

        >>> dr1.dataset.time.values
        array([ 0.])

        >>> dr1.variable_names
        ['mean_elevation']

        >>> dr1.dataset['mean_elevation'].values
        array([100])

        >>> dr1.dataset.attrs
        OrderedDict([('time_units', 'y')])

        Example of a DataRecord with item_id as the only dimension:
        >>> my_items2 = {'grid_element': np.array(('node', 'link'), dtype=str),
        ...              'element_id': np.array([1, 3])}
        >>> dr2 = DataRecord(grid,
        ...                  items=my_items2)

        Note that both arrays (grid_element and element_id) have 1 dimension
        as they only vary along the dimension 'item_id'.

        >>> dr2.dataset.to_dataframe()[['grid_element', 'element_id']]
                grid_element  element_id
        item_id
        0               node           1
        1               link           3

        Example of a DataRecord with dimensions time and item_id:
        >>> my_items3 = {'grid_element':np.array([['node'], ['link']]),
        ...              'element_id': np.array([[1], [3]])}
        >>> dr3 = DataRecord(grid,
        ...                  time=[0.],
        ...                  items=my_items3)

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.

        >>> dr3.dataset.to_dataframe()[['grid_element', 'element_id']]
                     grid_element  element_id
        item_id time
        0       0.0          node           1
        1       0.0          link           3
        """
        # save a reference to the grid
        self._grid = grid

        # depending on the grid type, permitted locations for items vary
        self._permitted_locations = self._grid.groups

        # set initial time coordinates, if any
        if isinstance(time, (list, np.ndarray)):
            self._times = np.array(time)
            self._number_of_times = len(self._times)
        elif time is not None:
            raise TypeError("time must be a list or numpy array")

        # set initial items, if any
        if items is not None:
            try:
                items.keys()
            except AttributeError:
                # items is not a dict
                raise TypeError(
                    "You must provide an `items` dictionary "
                    "(see documentation for required format)"
                )
            try:
                _grid_elements, _element_ids = (
                    items["grid_element"],
                    items["element_id"],
                )
            except KeyError:
                # grid_element and/or element_id not provided
                raise TypeError(
                    "You must provide an `items` dictionary,"
                    "(see documentation for required format)"
                )

            self._number_of_items = len(_element_ids)
            if len(_grid_elements) != self._number_of_items:
                if isinstance(_grid_elements, string_types):
                    pass
                else:
                    raise ValueError(
                        "The number of grid_element passed "
                        "to DataRecord must be 1 or equal "
                        "to the number of element_id."
                    )

            # check that grid_element and element_id exist on the grid and
            # have valid format:
            _grid_elements = self._check_grid_element_and_id(
                _grid_elements, _element_ids
            )

            # check that element IDs do not exceed number of elements
            # on the grid:
            self._check_element_id_values(_grid_elements, _element_ids)

            # create coordinates for the dimension 'item_id':
            self._item_ids = np.array(range(self._number_of_items))

            # create initial dictionaries of variables:
            if time is not None:
                data_vars_dict = {
                    "grid_element": (["item_id", "time"], _grid_elements),
                    "element_id": (["item_id", "time"], _element_ids),
                }
                coords = {"time": self._times, "item_id": self._item_ids}
            else:
                # no time
                data_vars_dict = {
                    "grid_element": (["item_id"], _grid_elements),
                    "element_id": (["item_id"], _element_ids),
                }
                coords = {"item_id": self._item_ids}

        else:
            # no items, initial dictionary of variables is empty:
            data_vars_dict = {}
            if time is not None:
                coords = {"time": self._times}
            else:  # no item and no time = no dimension
                coords = {}

        # set variables, if any
        if data_vars is not None:
            try:
                # check format (dict)
                data_vars.keys()
            except AttributeError:
                raise TypeError(
                    "Data variables (data_vars) passed to "
                    "DataRecord must be a dictionary (see "
                    "documentation for valid structure)"
                )
            for key in data_vars.keys():
                # check dict structure and dims:
                if data_vars[key][0] not in (
                    ["time"],
                    ["item_id"],
                    ["time", "item_id"],
                    ["item_id", "time"],
                ):
                    raise ValueError(
                        "Data variable dimensions must be " "time and/or item_id"
                    )

            # create complete dictionary of variables
            # (= initial data_vars_dict + additional user-defined data_vars):
            data_vars_dict.update(data_vars)

        # set attributes, if any
        if attrs is not None:
            try:
                attrs.keys()
            except AttributeError:
                raise TypeError(
                    "Attributes (attrs) passed to DataRecord" "must be a dictionary"
                )

        # create an xarray Dataset:
        self._dataset = Dataset(data_vars=data_vars_dict, coords=coords, attrs=attrs)

    def _check_grid_element_and_id(self, grid_element, element_id, flag=None):
        """Check the location and size of grid_element and element_id."""
        if isinstance(grid_element, string_types):
            # all items are on same type of grid_element
            if grid_element in self._permitted_locations:
                pass
            else:
                raise ValueError(
                    "Location provided: " + grid_element + " is "
                    "not a permitted location for this grid type"
                )

            # create list of grid_element for all items
            ge_name = grid_element
            try:
                # if time
                grid_element = np.array(
                    np.empty(
                        (self._number_of_items, self._number_of_times), dtype=object
                    )
                )
            except (AttributeError, RuntimeError):
                # no time
                grid_element = np.array(
                    np.empty((self._number_of_items,), dtype=object)
                )
            grid_element.fill(ge_name)

        else:
            # each item on different grid element
            for loc in grid_element:
                if isinstance(loc, np.ndarray):
                    # depending on dims
                    loc = loc[0]
                if loc in self._permitted_locations:
                    pass
                else:
                    raise ValueError(
                        "One or more of the grid elements"
                        " provided is/are not permitted location"
                        " for this grid type"
                    )

        return grid_element

    def _check_element_id_values(self, grid_element, element_id):
        """Check that element_id values are valid."""
        for at in self._permitted_locations:
            max_size = self._grid[at].size
            selected_elements_ind = [
                i for i in range(len(grid_element)) if grid_element[i] == at
            ]
            selected_elements = element_id[selected_elements_ind]

            if selected_elements.size > 0:
                if max(selected_elements) >= max_size:
                    raise ValueError(
                        "An item residing at " + at + " has an "
                        "element_id larger than the number of " + at + " on the grid"
                    )
                less_than_zero = selected_elements < 0
                if any(less_than_zero):
                    raise ValueError(
                        "An item residing at " + at + " has "
                        "an element id below zero. This is not "
                        "permitted"
                    )
        dtype = element_id.dtype
        if dtype != int:
            raise ValueError(
                "You have passed a non-integer element_id to "
                "DataRecord, this is not permitted"
            )

    def add_record(self, time=None, item_id=None, new_item_loc=None, new_record=None):
        """Add a time-related record to an existing DataRecord.

        Parameters
        ----------
        time : list or 1-D array of float or int
            Time step at which the record is to be added.
        item_id : list or 1-D array of int (optional)
            ID of the item to which the new record relates.
        new_item_loc: dict (optional)
            Dictionary of the new item location. If the new record is a change
            in the item location (grid_element and/or element_id), this field
            must be provided as:
                {'grid_element' : [grid_element],
                 'element_id' : [element_id]}
            Both must be provided even if only one is being changed.
        new_record : dict
            Dictionary containing the new record. Structure should be:
            {'variable_name_1' : (['dimensions'], variable_data_1)}
            with:
                - 'variable_name_1' : name of the (potentially new) variable
                - ['dimensions'] : dimension(s) along which the new record
                varies; can be ['time'], ['item_id] or ['item_id', 'time']
                - variable_data_1 : new data array, size must match the
                variable dimension(s)

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:
        >>> my_items3 = {'grid_element': np.array([['node'], ['link']]),
        ...              'element_id': np.array([[1],[3]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.
        >>> dr3 = DataRecord(grid,
        ...                  time=[0.],
        ...                  items=my_items3)

        Records relating to pre-existing items can be added to the DataRecord
        using the method 'add_record':
        >>> dr3.add_record(time=[2.0],
        ...                item_id=[0],
        ...                new_item_loc={'grid_element' : np.array([['node']]),
        ...                             'element_id' : np.array([[6]])},
        ...                new_record={'item_size':(
        ...                           ['item_id', 'time'], np.array([[0.2]]))})
        >>> dr3.dataset['element_id'].values
        array([[  1.,   6.],
               [  3.,  nan]])
        >>> dr3.get_data([2.0],[0],'item_size')
        array([ 0.2])

        The 'add_record' method can also be used to add a non item-related
        record:
        >>> dr3.add_record(time=[50.0],
        ...                new_record={'mean_elev': (['time'], [110])})
        >>> dr3.dataset['mean_elev'].to_dataframe()
              mean_elev
        time
        0.0         NaN
        2.0         NaN
        50.0      110.0
        """
        if time is not None:
            try:
                # check that time is a dim of the DataRecord
                self._dataset["time"]
            except KeyError:
                raise KeyError("This DataRecord does not record time")

            if not isinstance(time, (list, np.ndarray)):
                # check input type
                raise TypeError(
                    "You have passed a time that is"
                    " not permitted, must be list or array"
                )
            else:
                if item_id is not None:
                    try:
                        # check that DataRecord holds items
                        self._dataset["item_id"]
                    except KeyError:
                        raise KeyError("This DataRecord does not hold items")
                    try:
                        # check that item_id is list or array
                        len(item_id)
                    except TypeError:
                        raise TypeError("item_id must be a list or a 1D array")
                    if not all(i in self._dataset["item_id"].values for i in item_id):
                        # check that item_id already exist
                        raise ValueError(
                            "One or more of the value(s) you "
                            "passed as item_id is/are not "
                            "currently in the DataRecord. Change"
                            " the input values create a new item"
                            "using the method add_item"
                        )
                    coords_to_add = {"time": np.array(time), "item_id": item_id}

                    # if item location is changed by this new record, check
                    # that both grid_element and element_id are provided:
                    if new_item_loc is not None:
                        try:
                            new_grid_element = new_item_loc["grid_element"]
                            new_element_id = new_item_loc["element_id"]
                        except KeyError:
                            raise KeyError(
                                "You must provide a "
                                "new_item_loc dictionary with both "
                                "grid_element and element_id"
                            )
                        # check that grid_element and element_id exist
                        # on the grid and have valid format:
                        new_grid_element = self._check_grid_element_and_id(
                            new_grid_element, new_element_id, flag=1
                        )

                        # check that element IDs do not exceed number
                        # of elements on this grid:
                        self._check_element_id_values(new_grid_element, new_element_id)

                        _new_data_vars = {
                            "grid_element": (["item_id", "time"], new_grid_element),
                            "element_id": (["item_id", "time"], new_element_id),
                        }
                    else:
                        # new_item_loc is `None`
                        _new_data_vars = {}
                else:
                    # no item
                    coords_to_add = {"time": np.array(time)}
                    _new_data_vars = {}

        else:
            # no time
            if item_id is not None:
                if not all(i in self._dataset["item_id"].values for i in item_id):
                    # check that item_id already exist
                    raise ValueError(
                        "One or more of the value(s) you "
                        "passed as item_id is/are not "
                        "currently in the DataRecord. Change"
                        " the input values create a new item"
                        "using the method add_item"
                    )

                coords_to_add = {"item_id": np.array(item_id)}
                _new_data_vars = {}

                # no time so if item location needs to be changed,
                # user should use set_data
                if new_item_loc is not None:
                    raise ValueError(
                        "Use the method set_data to change the "
                        "location of an item in this DataRecord"
                    )
            else:
                # no item
                _new_data_vars = {}
                coords_to_add = {}

        if new_record is not None:
            # add new_record to dict of variables to add
            _new_data_vars.update(new_record)

        # create dataset of new record
        ds_to_add = Dataset(data_vars=_new_data_vars, coords=coords_to_add)
        # merge new record and original dataset
        self._dataset = self._dataset.merge(ds_to_add, compat="no_conflicts")

    def add_item(self, time=None, new_item=None, new_item_spec=None):
        """Add new item(s) to the current DataRecord.

        time : list or 1-D array of float or int
            Time step at which the items are to be added.
        new_item : dict
            Structure is:
                {'grid_element' : [grid_element],
                 'element_id' : [element_id]}
            where:
                - [grid_element] is str or number-of-items long array
                containing strings of the grid element(s) on which the items
                live. Valid locations depend on the grid type. If provided as
                a string it is assumed that all items live on the same type of
                grid element.
                - [element_id] is an array of integers identifying the grid
                element ID on which each item resides.
            An example argument would be:
                {'grid_element' : numpy.array([['node'], ['node'], ['link']]),
                 'element_id' :   numpy.array([[1],      [5],      [1]     ])}
        new_item_spec : dict (optional)
            Dictionary containing any data variables (other than
            'grid_element' and 'element_id') relating to the new item(s) to be
            added. Structure is:
                {'variable_name_1' : (['dimensions'], variable_data_1)}
            where:
                - 'variable_name_1' : name of the (potentially new) variable
                - ['dimensions'] : dimension(s) along which the new record
                varies; can be ['time'], ['item_id] or ['item_id', 'time']
                - variable_data_1 : new data array, size must match the
                variable dimension(s)

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:
        >>> my_items3 = {'grid_element':np.array([['node'], ['link']]),
        ...              'element_id': np.array([[1],[3]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.
        >>> dr3 = DataRecord(grid,
        ...                  time=[0.],
        ...                  items=my_items3)

        Items can be added to a DataRecord that already holds similar items,
        using the method 'add_item':
        >>> dr3.add_item(time=[1.0],
        ...              new_item={'grid_element' : np.array(
        ...                                              [['node'], ['node']]),
        ...                        'element_id' : np.array([[4],[4]])},
        ...              new_item_spec={'size': (
        ...                              ['item_id', 'time'], [[10],[5]])})

        Two items have been added at a new timestep 1.0:
        >>> dr3.number_of_items
        4
        >>> dr3.time_coordinates
        [0.0, 1.0]

        If a data variable is also added with the new items ('size' in this
        example), the values for this variable are filled with 'nan' for the
        pre-existing items:
        >>> dr3.dataset['size'][:,1].values
        array([ nan,  nan,  10.,   5.])

        The previous line calls the values of the variable 'size', for all
        items, at time=1; the first two items don't have a value for the
        variable 'size'.
        """
        if time is None and "time" in self._dataset["grid_element"].coords:
            raise ValueError(
                "The items previously defined in this DataRecord"
                ' have dimensions "time" and "item_id", '
                'you must provide a "time" for the new item(s)'
            )

        try:
            # check that new_item is a dict
            new_item.keys()
        except AttributeError:
            raise TypeError(
                "You must provide an new_item dictionary "
                "(see documentation for required format)"
            )
        try:
            # check that dict contains correct entries
            _grid_elements, _element_ids = (
                new_item["grid_element"],
                new_item["element_id"],
            )
        except KeyError:
            raise KeyError(
                "You must provide a new_item dictionary "
                "(see documentation for required format)"
            )

        number_of_new_items = len(new_item["element_id"])
        # first id of new item = last item in existing datarecord+1
        new_first_item_id = self._dataset["item_id"][-1].values + 1
        new_item_ids = np.array(
            range(new_first_item_id, new_first_item_id + number_of_new_items)
        )

        if time is not None:
            try:
                self._dataset["time"]
            except KeyError:
                raise KeyError("This DataRecord does not record time")
            if not isinstance(time, (list, np.ndarray)):
                raise TypeError(
                    "You have passed a time that is not "
                    "permitted, must be list or a 1-D array"
                )
            else:
                coords_to_add = {
                    "time": np.array(time),
                    "item_id": np.array(new_item_ids),
                }
                # check that grid_element and element_id exist
                # on the grid and have valid format
                _grid_elements = self._check_grid_element_and_id(
                    _grid_elements, _element_ids
                )

                # check that element IDs do not exceed number
                # of elements on this grid
                self._check_element_id_values(_grid_elements, _element_ids)

                data_vars_dict = {
                    "grid_element": (["item_id", "time"], _grid_elements),
                    "element_id": (["item_id", "time"], _element_ids),
                }

        else:
            # no time
            coords_to_add = {"item_id": np.array(new_item_ids)}
            # check that grid_element and element_id exist on
            # the grid and have valid format:
            _grid_elements = self._check_grid_element_and_id(
                _grid_elements, _element_ids
            )
            # check that element IDs do not exceed number of
            # elements on this grid
            self._check_element_id_values(_grid_elements, _element_ids)

            data_vars_dict = {
                "grid_element": (["item_id"], _grid_elements),
                "element_id": (["item_id"], _element_ids),
            }

        # other variables:
        if new_item_spec is not None:
            data_vars_dict.update(new_item_spec)

        # Dataset of new record:
        ds_to_add = Dataset(data_vars=data_vars_dict, coords=coords_to_add)

        # Merge new record and original dataset:
        self._dataset = self._dataset.merge(ds_to_add, compat="no_conflicts")

    def get_data(self, time=None, item_id=None, data_variable=None):
        """Get the value of a variable at a model time and/or for an item.

        Parameters
        ----------
        time : list or 1-D array of float or int (optional)
            The time coordinate of the record to get.
        item_id : list or 1-D array of int (optional)
            The item id of the record to get.
        data_variable : string
            The label of the variable to get.

        Returns
        -------
        object
            The value of *variable* at *time* and/or for *item_id*. The type of
            the returned object is dependent on the type of the variable value.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:
        >>> my_items4 = {'grid_element' : 'node',
        ...              'element_id': np.array([[1],[3],[3],[7]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.
        >>> my_data4 = {'item_size': (['item_id', 'time'], np.array(
        ...                                 [[0.3],[0.4],[0.8],[0.4]]))}
        >>> dr4=DataRecord(grid,
        ...                time=[50.],
        ...                items=my_items4,
        ...                data_vars=my_data4)
        >>> dr4.get_data([50.],[2],'element_id')
        array([3])
        >>> dr4.get_data(time=[50.],data_variable='item_size')
        array([  0.3,  0.4,  0.8,  0.4])
        >>> dr4.get_data(item_id=[1,2], data_variable='grid_element')
        array([['node'],
               ['node']], dtype=object)
        """
        try:
            self._dataset[data_variable]
        except KeyError:
            raise KeyError(
                "the variable '{}' is not in the " "DataRecord".format(data_variable)
            )
        if time is None:
            if item_id is None:
                return self._dataset[data_variable].values
            else:
                try:
                    self._dataset["item_id"]
                except KeyError:
                    raise KeyError("This DataRecord does not hold items")
                try:
                    len(item_id)
                except TypeError:
                    raise TypeError("item_id must be a list or a 1-D array")
                try:
                    self._dataset["item_id"].values[item_id]
                except IndexError:
                    raise IndexError(
                        "The item_id you passed does not exist " "in this DataRecord"
                    )

                return self._dataset.isel(item_id=item_id)[data_variable].values

        else:  # time is not None
            try:
                self._dataset["time"]
            except KeyError:
                raise KeyError("This DataRecord does not record time")
            try:
                len(time)
            except TypeError:
                raise TypeError("time must be a list or a 1-D array")
            try:
                time_index = int(self.time_coordinates.index(time[0]))
            except ValueError:
                raise IndexError(
                    "The time you passed is not currently"
                    " in the DataRecord, you must change the value"
                    " you pass or first create the new time "
                    " coordinate using the add_record method"
                )
            if item_id is None:
                return self._dataset.isel(time=time_index)[data_variable].values
            else:
                try:
                    self._dataset["item_id"]
                except KeyError:
                    raise KeyError("This DataRecord does not hold items")
                try:
                    len(item_id)
                except TypeError:
                    raise TypeError("item_id must be a list or a 1-D array")
                try:
                    self._dataset["item_id"].values[item_id]
                except IndexError:
                    raise IndexError(
                        "The item_id you passed does not exist " "in this DataRecord"
                    )
                return self._dataset.isel(time=time_index, item_id=item_id)[
                    data_variable
                ].values

    def set_data(self, time=None, item_id=None, data_variable=None, new_value=np.nan):
        """Set a variable value at a model time and/or an item to a new value.

        The value of only one variable can be changed at a time using this
        method.

        Parameters
        ----------
        time : list or 1-D array of float or int
            The time coordinate of the record to set.
        item_id : list or 1-D array of int
            The item id of the record to set.
        data_variable : string
            The label of the variable to set.
        new_value : list or 1-D array
            The new value to give to the variable data.

        Returns
        -------
        DataRecord with updated data.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:
        >>> my_items4 = {'grid_element' : 'node',
        ...              'element_id': np.array([[1],[3],[3],[7]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.
        >>> my_data4 = {'item_size': (['item_id', 'time'], np.array(
        ...                                 [[0.3],[0.4],[0.8],[0.4]]))}
        >>> dr4 = DataRecord(grid,
        ...                  time=[50.],
        ...                  items=my_items4,
        ...                  data_vars=my_data4)
        >>> dr4.dataset['item_size'].values
        array([[ 0.3],
               [ 0.4],
               [ 0.8],
               [ 0.4]])
        >>> dr4.set_data([50.],[2],'item_size', [0.5])
        >>> dr4.dataset['item_size'].values
        array([[ 0.3],
               [ 0.4],
               [ 0.5],
               [ 0.4]])
        """
        if data_variable not in self.variable_names:
            raise KeyError(
                "the variable '{}' is not in the " "DataRecord".format(data_variable)
            )

        # If record to be changed is 'grid_element' or 'element_id',
        # check that provided grid_element is valid and that new
        # grid_element+element_id combination exist on the grid and
        # have valid format:
        if data_variable in ("grid_element", "element_id"):
            if data_variable == "grid_element":
                assoc_grid_element = new_value
                assoc_element_id = self.get_data(time, item_id, "element_id")[0]
            if data_variable == "element_id":
                if not isinstance(new_value, int):
                    raise ValueError(
                        "You have passed a non-integer "
                        "element_id to DataRecord, this is not "
                        "permitted"
                    )
                if new_value < 0:
                    raise ValueError(
                        "You have passed an element id below "
                        "zero. This is not permitted"
                    )
                assoc_element_id = new_value
                assoc_grid_element = self.get_data(time, item_id, "grid_element")[0]
            self._check_grid_element_and_id(
                assoc_grid_element, assoc_element_id, flag=1
            )
            if assoc_element_id >= self._grid[assoc_grid_element].size:
                raise ValueError(
                    "The location "
                    + assoc_grid_element
                    + " "
                    + str(assoc_element_id)
                    + " does not exist on this grid"
                )

        if time is None:
            self._dataset[data_variable].values[item_id] = new_value
        else:
            try:
                len(time)
            except TypeError:
                raise TypeError("time must be a list or a 1-d array")
            try:
                # check that time coordinate already exists
                time_index = np.where(self._dataset.time.values == time)[0][0]
            except IndexError:
                raise IndexError(
                    "The time you passed is not currently"
                    " in the DataRecord, you must change the value"
                    " you pass or first create the new time "
                    " coordinate using the add_record method"
                )

            if item_id is None:
                self._dataset[data_variable].values[time_index] = new_value
            else:
                try:
                    len(item_id)
                except TypeError:
                    raise TypeError("item_id must be a list or a 1-d array")
                try:
                    self._dataset["item_id"]
                    self._dataset[data_variable].values[item_id, time_index] = new_value
                except KeyError:
                    raise KeyError("This DataRecord does not hold items")

    def calc_aggregate_value(
        self, func, data_variable, at="node", filter_array=None, args=(), **kwargs
    ):
        """Apply a function to a variable aggregated at grid elements.

        Parameters
        ----------
        func : function
            Function to apply to be aggregated.
        data_variable : str
            Name of variable on which to apply the function.
        at : str, optional
            Name of grid element at which to apply the function.
            Default is "node".
        filter_array: boolean array with dimensions matching that of the
            DataRecord (optional)
            Array to filter the DataRecord before aggregation.
        args : tuple (optional)
            Additional positional arguments to pass to the function.
        **kwargs : key value pairs (optional)
            Additional keyword arguments to pass to func.

        Returns
        -------
        out : ndarray
            Array of size number-of-grid_elements (grid_elements is the group
            passed as 'at' argument).

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.data_record import DataRecord
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((3,3))
        >>> element_id = [0, 0, 0, 0, 1, 2, 3, 4, 5]
        >>> volumes = [4, 5, 1, 2, 3, 4, 5, 6, 7]
        >>> ages = [10, 11, 12, 13, 14, 15, 16, 8, 10]
        >>> grid_element = 'node'
        >>> data = {'ages': ages,
        ...         'volumes': volumes}
        >>> dr = DataRecord(grid,
        ...                 items={'grid_element' : 'node',
        ...                           'element_id' : np.array(element_id)},
        ...                 data_vars={'ages' : (['item_id'], np.array(ages)),
        ...                             'volumes' : (
        ...                                 ['item_id'], np.array(volumes))})
        >>> s = dr.calc_aggregate_value(func=np.sum, data_variable='ages')
        >>> s
        array([ 46.,  14.,  15.,  16.,   8.,  10.,  nan,  nan,  nan])
        >>> len(s) == grid.number_of_nodes
        True

        If you want to first filter the DataRecord and then aggregate, first
        create a filter array with dimensions matching that of the DataRecord
        and has `True` for entries that should be retained and False for
        entries that should be ignored.

        For example, if we wanted to aggregate volume for items with an age
        greater than 10 we would to the following:
        >>> f = dr.dataset['ages'] > 10.
        >>> v_f = dr.calc_aggregate_value(func=np.sum,
        ...                               data_variable='volumes',
        ...                               filter_array=f)
        >>> v_f
        array([  8.,   3.,   4.,   5.,  nan,  nan,  nan,  nan,  nan])

        """
        filter_at = self._dataset["grid_element"] == at

        if filter_array is None:
            my_filter = filter_at
        else:
            my_filter = filter_at & filter_array

        # Filter DataRecord with my_filter and groupby element_id:
        filtered = self._dataset.where(my_filter).groupby("element_id")

        vals = filtered.apply(func, *args, **kwargs)  # .reduce

        # create a nan array that we will fill with the results of the sum
        # this should be the size of the number of elements, even if there are
        # no items living at some grid elements.
        out = np.nan * np.ones(self._grid[at].size)

        # put the values of the specified variable into the correct location
        # of the out array.
        out[vals.element_id.values.astype(int)] = vals[data_variable]

        return out

    def ffill_grid_element_and_id(self):
        """Fill NaN values of the fields 'grid_element' and 'element_id'.

        Fields are filled by propagating values forward in time.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.data_record import DataRecord
        >>> grid = RasterModelGrid((3,3))

        Example of a DataRecord with dimensions time and item_id:
        >>> my_items3 = {'grid_element': np.array([['node'], ['link']]),
        ...              'element_id': np.array([[1],[3]])}

        Note that both arrays have 2 dimensions as they vary along dimensions
        'time' and 'item_id'.
        >>> dr3 = DataRecord(grid,
        ...                  time=[0.],
        ...                  items=my_items3)

        Records relating to pre-existing items can be added to the DataRecord
        using the method 'add_record':
        >>> dr3.add_record(time=[2.0, 3.0],
        ...                new_record={'mean_elevation':(
        ...                           ['time'], np.array([200., 250.]))})

        Adding this data record created two new time coordinates. The
        grid_element and element_id of the items has been filled with 'nan'
        for these time coordinates.

        >>> dr3.dataset['grid_element'].values
        array([['node', nan, nan],
               ['link', nan, nan]], dtype=object)
        >>> dr3.dataset['element_id'].values
        array([[  1.,  nan,  nan],
               [  3.,  nan,  nan]])

        To fill these values with the last valid value, use the method
        ffill_grid_element_and_id:
        >>> dr3.ffill_grid_element_and_id()
        >>> dr3.dataset['grid_element'].values
        array([['node', 'node', 'node'],
               ['link', 'link', 'link']], dtype=object)
        >>> dr3.dataset['element_id'].values
        array([[ 1.,  1.,  1.],
               [ 3.,  3.,  3.]])

        """
        # Forward fill element_id:
        fill_value = []
        ei = self._dataset["element_id"].values
        for i in range(ei.shape[0]):
            for j in range(ei.shape[1]):
                if np.isnan(ei[i, j]):
                    ei[i, j] = fill_value
                else:
                    fill_value = ei[i, j]
        self._dataset["element_id"] = (["item_id", "time"], ei)
        # Can't do ffill to grid_element because str/nan, so:
        fill_value = ""
        ge = self._dataset["grid_element"].values
        for i in range(ge.shape[0]):
            for j in range(ge.shape[1]):
                if isinstance(ge[i, j], str):
                    fill_value = ge[i, j]
                else:
                    ge[i, j] = fill_value
        self._dataset["grid_element"] = (["item_id", "time"], ge)

    @property
    def dataset(self):
        """The xarray Dataset that serves as the core datastructure.
        """
        return self._dataset

    @property
    def variable_names(self):
        """Return the name(s) of the data variable(s) in the record as a list.
        """
        _keys = []
        for key in self._dataset.to_dataframe().keys():
            _keys.append(key)
        return _keys

    @property
    def number_of_items(self):
        """Return the number of items in the DataRecord.
        """
        return len(self._dataset.item_id)

    @property
    def item_coordinates(self):
        """Return a list of the item_id coordinates in the DataRecord.
        """
        return self._dataset.item_id.values.tolist()

    @property
    def number_of_timesteps(self):
        """Return the number of time steps in the DataRecord.
        """
        return len(self._dataset.time)

    @property
    def time_coordinates(self):
        """Return a list of the time coordinates in the DataRecord.
        """
        return self._dataset.time.values.tolist()

    @property
    def earliest_time(self):
        """Return the earliest time coordinate in the DataRecord.
        """
        return min(self._dataset.time.values)

    @property
    def latest_time(self):
        """Return the latest time coordinate in the DataRecord.
        """
        return max(self._dataset.time.values)

    @property
    def prior_time(self):
        """Return the penultimate time coordinate in the DataRecord.
        """
        if self.number_of_timesteps < 2:
            return np.nan
        else:
            return sorted(self.time_coordinates)[-2]



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.soil_moisture.SoilInfiltrationGreenAmpt

last updated: 3/14/16
"""
import numpy as np

from landlab import RasterModelGrid
from landlab.components.soil_moisture import SoilInfiltrationGreenAmpt

(_SHAPE, _SPACING, _ORIGIN) = ((10, 10), (25, 25), (0.0, 0.0))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_SI_name(si):
    assert si.name == "SoilInfiltrationGreenAmpt"


def test_SI_input_var_names(si):
    assert si.input_var_names == (
        "surface_water__depth",
        "soil_water_infiltration__depth",
    )


def test_SI_output_var_names(si):
    assert si.output_var_names == (
        "surface_water__depth",
        "soil_water_infiltration__depth",
    )


def test_SI_var_units(si):
    assert set(si.input_var_names) | set(si.output_var_names) == set(
        dict(si.units).keys()
    )

    assert si.var_units("surface_water__depth") == "m"
    assert si.var_units("soil_water_infiltration__depth") == "m"


def test_grid_shape(si):
    assert si.grid.number_of_node_rows == _SHAPE[0]
    assert si.grid.number_of_node_columns == _SHAPE[1]


def test_calc_soil_pressure(si):
    np.testing.assert_almost_equal(
        si.calc_soil_pressure("silt loam"), 0.1647870740305523, decimal=6
    )


def test_calc_soil_head(si):
    soil_props = SoilInfiltrationGreenAmpt.SOIL_PROPS["loam"]
    np.testing.assert_almost_equal(
        si.calc_pressure_head(soil_props[0], soil_props[1]), 0.087498292, decimal=6
    )


def test_calc_moisture_deficit(si):
    np.testing.assert_almost_equal(
        si.calc_moisture_deficit(
            soil_bulk_density=1700.0,
            rock_density=2650.0,
            volume_fraction_coarse_fragments=0.0,
            soil_moisture_content=0.2,
        ),
        0.15849056603,
        decimal=6,
    )


def test_run_one_step():
    grid = RasterModelGrid((10, 10), xy_spacing=25)
    grid.add_ones("node", "soil_water_infiltration__depth", dtype=float)
    grid.add_ones("node", "surface_water__depth")
    hydraulic_conductivity = 2.5 * (10 ** -6)
    grid["node"]["surface_water__depth"] *= 5.0
    grid["node"]["soil_water_infiltration__depth"] *= 10 ** -5
    SI = SoilInfiltrationGreenAmpt(
        grid,
        hydraulic_conductivity=hydraulic_conductivity,
        soil_bulk_density=1700.0,
        rock_density=2650.0,
        initial_soil_moisture_content=0.2,
        soil_type="silt loam",
        volume_fraction_coarse_fragments=0.6,
        coarse_sed_flag=False,
        surface_water_minimum_depth=1.0e-7,
        soil_pore_size_distribution_index=None,
        soil_bubbling_pressure=None,
        wetting_front_capillary_pressure_head=None,
    )

    SI.run_one_step(dt=5)
    np.testing.assert_almost_equal(
        grid["node"]["surface_water__depth"][0], 3.97677483519, decimal=6
    )



#! /usr/bin/env python
"""
Unit tests for landlab.io.esri_ascii module.
"""
import os

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal
from six import StringIO

from landlab import RasterModelGrid
from landlab.io import (
    BadHeaderLineError,
    DataSizeError,
    KeyTypeError,
    KeyValueError,
    MismatchGridDataSizeError,
    MismatchGridXYLowerLeft,
    MismatchGridXYSpacing,
    MissingRequiredKeyError,
    read_asc_header,
    read_esri_ascii,
)

_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


def test_hugo_read_file_name():
    (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR, "hugo_site.asc"))

    assert isinstance(grid, RasterModelGrid)

    assert field.size == 55 * 76
    assert field.shape == (55 * 76,)
    assert (grid.x_of_node.min(), grid.y_of_node.min()) == (0.0, 0.0)


def test_hugo_read_file_like():
    with open(os.path.join(_TEST_DATA_DIR, "hugo_site.asc")) as asc_file:
        (grid, field) = read_esri_ascii(asc_file)

    assert isinstance(grid, RasterModelGrid)

    assert field.size == 55 * 76
    assert field.shape == (55 * 76,)


def test_hugo_reshape():
    with open(os.path.join(_TEST_DATA_DIR, "hugo_site.asc")) as asc_file:
        (grid, field) = read_esri_ascii(asc_file, reshape=True)

    assert isinstance(grid, RasterModelGrid)

    assert field.shape == (55, 76)


def test_4x3_read_file_name():
    (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR, "4_x_3.asc"))

    assert isinstance(grid, RasterModelGrid)

    assert isinstance(field, np.ndarray)
    assert (grid.x_of_node.min(), grid.y_of_node.min()) == (1.0, 2.0)
    assert_array_equal(
        field, np.array([9.0, 10.0, 11.0, 6.0, 7.0, 8.0, 3.0, 4.0, 5.0, 0.0, 1.0, 2.0])
    )


def test_4x3_read_file_like():
    with open(os.path.join(_TEST_DATA_DIR, "4_x_3.asc")) as asc_file:
        (grid, field) = read_esri_ascii(asc_file)

    assert isinstance(grid, RasterModelGrid)

    assert_array_equal(
        field, np.array([9.0, 10.0, 11.0, 6.0, 7.0, 8.0, 3.0, 4.0, 5.0, 0.0, 1.0, 2.0])
    )


def test_4x3_shape_mismatch():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4.
5. 6. 7. 8.
9. 10. 11. 12.
        """
    )
    (grid, field) = read_esri_ascii(asc_file)
    assert field.size == 12

    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10. 11. 12.
        """
    )
    (grid, field) = read_esri_ascii(asc_file)
    assert field.size == 12


def test_4x3_size_mismatch():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10.
        """
    )
    with pytest.raises(DataSizeError):
        read_esri_ascii(asc_file)


def test_4x3_size_mismatch_with_halo():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10.
        """
    )
    with pytest.raises(DataSizeError):
        read_esri_ascii(asc_file, halo=1)


def test_grid_data_size_mismatch():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10. 11. 12.
        """
    )
    rmg = RasterModelGrid((10, 10), xy_spacing=10.0)
    with pytest.raises(MismatchGridDataSizeError):
        read_esri_ascii(asc_file, grid=rmg)


def test_grid_dx_mismatch():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10. 11. 12.
        """
    )
    rmg = RasterModelGrid((4, 3), xy_spacing=15.0)
    with pytest.raises(MismatchGridXYSpacing):
        read_esri_ascii(asc_file, grid=rmg)


def test_grid_lower_left_mismatch():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10. 11. 12.
        """
    )
    rmg = RasterModelGrid((4, 3), xy_spacing=10.0, xy_of_lower_left=(10, 15))
    with pytest.raises(MismatchGridXYLowerLeft):
        read_esri_ascii(asc_file, grid=rmg)


def test_header_missing_required_key():
    asc_file = StringIO(
        """
nrows         4
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """
    )
    with pytest.raises(MissingRequiredKeyError):
        read_asc_header(asc_file)


def test_header_unknown_key():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
invalid_key   1
        """
    )
    with pytest.raises(BadHeaderLineError):
        read_asc_header(asc_file)


def test_header_missing_value():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize
NODATA_value  -9999
invalid_key   1
        """
    )
    with pytest.raises(BadHeaderLineError):
        read_asc_header(asc_file)


def test_header_bad_values():
    asc_file = StringIO(
        """
nrows         -4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """
    )
    with pytest.raises(KeyValueError):
        read_asc_header(asc_file)


def test_header_missing_mutex_key():
    asc_file = StringIO(
        """
ncols         3
nrows         4
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """
    )
    with pytest.raises(MissingRequiredKeyError):
        read_asc_header(asc_file)


def test_header_mutex_key():
    asc_file = StringIO(
        """
ncols         3
nrows         4
xllcenter     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """
    )
    header = read_asc_header(asc_file)
    assert header["xllcenter"] == 1.0
    with pytest.raises(KeyError):
        header["xllcorner"]

    asc_file = StringIO(
        """
ncols         3
nrows         4
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """
    )
    header = read_asc_header(asc_file)
    assert header["xllcorner"] == 1.0
    with pytest.raises(KeyError):
        header["xllcenter"]


def test_header_missing_optional():
    asc_file = StringIO(
        """
ncols         3
nrows         4
xllcenter     1.
yllcorner     2.
cellsize      10.
        """
    )
    header = read_asc_header(asc_file)
    with pytest.raises(KeyError):
        header["nodata_value"]


def test_header_case_insensitive():
    asc_file = StringIO(
        """
nCoLs         3
nrows         4
Xllcenter     1.
YLLCORNER     2.
CELLSIZE      10.
NODATA_value  -999
        """
    )
    header = read_asc_header(asc_file)
    for key in ["ncols", "nrows", "xllcenter", "yllcorner", "cellsize", "nodata_value"]:
        assert key in header


def test_header_wrong_type():
    asc_file = StringIO(
        """
nCoLs         3.5
nrows         4
Xllcenter     1.
YLLCORNER     2.
CELLSIZE      10.
NODATA_value  -999
        """
    )
    with pytest.raises(KeyTypeError):
        read_asc_header(asc_file)


def test_name_keyword():
    (grid, field) = read_esri_ascii(
        os.path.join(_TEST_DATA_DIR, "4_x_3.asc"), name="air__temperature"
    )

    assert isinstance(grid, RasterModelGrid)

    assert isinstance(field, np.ndarray)
    assert_array_equal(
        field, np.array([9.0, 10.0, 11.0, 6.0, 7.0, 8.0, 3.0, 4.0, 5.0, 0.0, 1.0, 2.0])
    )
    assert_array_almost_equal(grid.at_node["air__temperature"], field)
    assert grid.at_node["air__temperature"] is field


def test_halo_keyword():
    (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR, "4_x_3.asc"), halo=1)

    assert isinstance(grid, RasterModelGrid)

    assert isinstance(field, np.ndarray)
    assert_array_equal(
        field,
        np.array(
            [
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                9.0,
                10.0,
                11.0,
                -9999.0,
                -9999.0,
                6.0,
                7.0,
                8.0,
                -9999.0,
                -9999.0,
                3.0,
                4.0,
                5.0,
                -9999.0,
                -9999.0,
                0.0,
                1.0,
                2.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
            ]
        ),
    )


def test_halo_keyword_no_nodata_value():
    (grid, field) = read_esri_ascii(
        os.path.join(_TEST_DATA_DIR, "4_x_3_no_nodata_value.asc"), halo=1
    )

    assert isinstance(grid, RasterModelGrid)

    assert isinstance(field, np.ndarray)
    assert_array_equal(
        field,
        np.array(
            [
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                9.0,
                10.0,
                11.0,
                -9999.0,
                -9999.0,
                6.0,
                7.0,
                8.0,
                -9999.0,
                -9999.0,
                3.0,
                4.0,
                5.0,
                -9999.0,
                -9999.0,
                0.0,
                1.0,
                2.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
                -9999.0,
            ]
        ),
    )



import numpy as np

from landlab import Component

from ...utils.decorators import use_file_name_or_kwds

_VALID_METHODS = set(["Grid"])


def assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError("%s: Invalid method name" % method)


class Vegetation(Component):
    """
    Landlab component that simulates net primary productivity, biomass
    and leaf area index at each cell based on inputs of root-zone
    average soil moisture.

    Ref: Zhou, X., Istanbulluoglu, E., & Vivoni, E. R. (2013). Modeling the
    ecohydrological role of aspect controlled radiation on tree grass shrub
    coexistence in a semiarid climate. Water Resources Research,
    49(5), 2872-2895.

    .. codeauthor:: Sai Nudurupati and Erkan Istanbulluoglu

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import Vegetation

    Create a grid on which to simulate vegetation dynamics.

    >>> grid = RasterModelGrid((5,4), xy_spacing=(0.2, 0.2))

    The grid will need some input data. To check the names of the fields
    that provide the input to this component, use the *input_var_names*
    class property.

    >>> sorted(Vegetation.input_var_names)  # doctest: +NORMALIZE_WHITESPACE
    ['surface__evapotranspiration',
     'surface__potential_evapotranspiration_30day_mean',
     'surface__potential_evapotranspiration_rate',
     'vegetation__plant_functional_type',
     'vegetation__water_stress']

    >>> sorted(Vegetation.units) # doctest: +NORMALIZE_WHITESPACE
    [('surface__evapotranspiration', 'mm'),
     ('surface__potential_evapotranspiration_30day_mean', 'mm'),
     ('surface__potential_evapotranspiration_rate', 'mm'),
     ('vegetation__cover_fraction', 'None'),
     ('vegetation__dead_biomass', 'g m^-2 d^-1'),
     ('vegetation__dead_leaf_area_index', 'None'),
     ('vegetation__live_biomass', 'g m^-2 d^-1'),
     ('vegetation__live_leaf_area_index', 'None'),
     ('vegetation__plant_functional_type', 'None'),
     ('vegetation__water_stress', 'None')]

    Provide all the input fields.

    >>> grid['cell']['vegetation__plant_functional_type']= (
    ...         np.zeros(grid.number_of_cells, dtype=int))
    >>> grid['cell']['surface__evapotranspiration'] = (
    ...         0.2 * np.ones(grid.number_of_cells))
    >>> grid['cell']['surface__potential_evapotranspiration_rate']= np.array([
    ...         0.25547770, 0.25547770, 0.22110221,
    ...         0.22110221, 0.24813062, 0.24813062])
    >>> grid['cell']['surface__potential_evapotranspiration_30day_mean']= (
    ...        np.array([0.25547770, 0.25547770, 0.22110221,
    ...                  0.22110221, 0.24813062, 0.24813062]))
    >>> grid['cell']['vegetation__water_stress'] = (
    ...        0.01 * np.ones(grid.number_of_cells))

    Instantiate the 'Vegetation' component.

    >>> Veg = Vegetation(grid)

    >>> Veg.grid.number_of_cell_rows
    3
    >>> Veg.grid.number_of_cell_columns
    2
    >>> Veg.grid is grid
    True
    >>> import numpy as np
    >>> sorted(Vegetation.output_var_names)  # doctest: +NORMALIZE_WHITESPACE
    ['vegetation__cover_fraction',
     'vegetation__dead_biomass',
     'vegetation__dead_leaf_area_index',
     'vegetation__live_biomass',
     'vegetation__live_leaf_area_index']

    >>> np.all(grid.at_cell['vegetation__live_leaf_area_index'] == 0.)
    True

    >>> Veg.update()

    >>> np.all(grid.at_cell['vegetation__live_leaf_area_index'] == 0.)
    False
    """

    _name = "Vegetation"

    _input_var_names = (
        "surface__evapotranspiration",
        "vegetation__water_stress",
        "surface__potential_evapotranspiration_rate",
        "surface__potential_evapotranspiration_30day_mean",
        "vegetation__plant_functional_type",
    )

    _output_var_names = (
        "vegetation__live_leaf_area_index",
        "vegetation__dead_leaf_area_index",
        "vegetation__cover_fraction",
        "vegetation__live_biomass",
        "vegetation__dead_biomass",
    )

    _var_units = {
        "vegetation__live_leaf_area_index": "None",
        "vegetation__dead_leaf_area_index": "None",
        "vegetation__cover_fraction": "None",
        "surface__evapotranspiration": "mm",
        "surface__potential_evapotranspiration_rate": "mm",
        "surface__potential_evapotranspiration_30day_mean": "mm",
        "vegetation__water_stress": "None",
        "vegetation__live_biomass": "g m^-2 d^-1",
        "vegetation__dead_biomass": "g m^-2 d^-1",
        "vegetation__plant_functional_type": "None",
    }

    _var_mapping = {
        "vegetation__live_leaf_area_index": "cell",
        "vegetation__dead_leaf_area_index": "cell",
        "vegetation__cover_fraction": "cell",
        "surface__evapotranspiration": "cell",
        "surface__potential_evapotranspiration_rate": "cell",
        "surface__potential_evapotranspiration_30day_mean": "cell",
        "vegetation__water_stress": "cell",
        "vegetation__live_biomass": "cell",
        "vegetation__dead_biomass": "cell",
        "vegetation__plant_functional_type": "cell",
    }

    _var_doc = {
        "vegetation__live_leaf_area_index": "one-sided green leaf area per unit ground surface area",
        "vegetation__dead_leaf_area_index": "one-sided dead leaf area per unit ground surface area",
        "vegetation__cover_fraction": "fraction of land covered by vegetation",
        "surface__evapotranspiration": "actual sum of evaporation and plant transpiration",
        "surface__potential_evapotranspiration_rate": "potential sum of evaporation and platnt transpiration",
        "surface__potential_evapotranspiration_30day_mean": "30 day mean of surface__potential_evapotranspiration",
        "vegetation__water_stress": "parameter that represents nonlinear effects of water defecit \
             on plants",
        "vegetation__live_biomass": "weight of green organic mass per unit area - measured in terms \
             of dry matter",
        "vegetation__dead_biomass": "weight of dead organic mass per unit area - measured in terms \
             of dry matter",
        "vegetation__plant_functional_type": "classification of plants (int), grass=0, shrub=1, tree=2, \
             bare=3, shrub_seedling=4, tree_seedling=5",
    }

    @use_file_name_or_kwds
    def __init__(
        self,
        grid,
        Blive_init=102.0,
        Bdead_init=450.0,
        ETthreshold_up=3.8,
        ETthreshold_down=6.8,
        Tdmax=10.0,
        w=0.55,
        WUE_grass=0.01,
        LAI_max_grass=2.0,
        cb_grass=0.0047,
        cd_grass=0.009,
        ksg_grass=0.012,
        kdd_grass=0.013,
        kws_grass=0.02,
        WUE_shrub=0.0025,
        LAI_max_shrub=2.0,
        cb_shrub=0.004,
        cd_shrub=0.01,
        ksg_shrub=0.002,
        kdd_shrub=0.013,
        kws_shrub=0.02,
        WUE_tree=0.0045,
        LAI_max_tree=4.0,
        cb_tree=0.004,
        cd_tree=0.01,
        ksg_tree=0.002,
        kdd_tree=0.013,
        kws_tree=0.01,
        WUE_bare=0.01,
        LAI_max_bare=0.01,
        cb_bare=0.0047,
        cd_bare=0.009,
        ksg_bare=0.012,
        kdd_bare=0.013,
        kws_bare=0.02,
        **kwds
    ):
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        Blive_init: float, optional
            Initial value for vegetation__live_biomass. Converted to field.
        Bdead_init: float, optional
            Initial value for vegetation__dead_biomass. Coverted to field.
        ETthreshold_up: float, optional
            Potential Evapotranspiration (PET) threshold for
            growing season (mm/d).
        ETthreshold_down: float, optional
            PET threshold for dormant season (mm/d).
        Tdmax: float, optional
            Constant for dead biomass loss adjustment (mm/d).
        w: float, optional
            Conversion factor of CO2 to dry biomass (Kg DM/Kg CO2).
        WUE: float, optional
            Water Use Efficiency - ratio of water used in plant water
            lost by the plant through transpiration (KgCO2Kg-1H2O).
        LAI_max: float, optional
            Maximum leaf area index (m2/m2).
        cb: float, optional
            Specific leaf area for green/live biomass (m2 leaf g-1 DM).
        cd: float, optional
            Specific leaf area for dead biomass (m2 leaf g-1 DM).
        ksg: float, optional
            Senescence coefficient of green/live biomass (d-1).
        kdd: float, optional
            Decay coefficient of aboveground dead biomass (d-1).
        kws: float, optional
            Maximum drought induced foliage loss rate (d-1).
        """
        self._method = kwds.pop("method", "Grid")

        assert_method_is_valid(self._method)

        super(Vegetation, self).__init__(grid)

        self.initialize(
            Blive_init=Blive_init,
            Bdead_init=Bdead_init,
            ETthreshold_up=ETthreshold_up,
            ETthreshold_down=ETthreshold_down,
            Tdmax=Tdmax,
            w=w,
            WUE_grass=WUE_grass,
            LAI_max_grass=LAI_max_grass,
            cb_grass=cb_grass,
            cd_grass=cd_grass,
            ksg_grass=ksg_grass,
            kdd_grass=kdd_grass,
            kws_grass=kws_grass,
            WUE_shrub=WUE_shrub,
            LAI_max_shrub=LAI_max_shrub,
            cb_shrub=cb_shrub,
            cd_shrub=cd_shrub,
            ksg_shrub=ksg_shrub,
            kdd_shrub=kdd_shrub,
            kws_shrub=kws_shrub,
            WUE_tree=WUE_tree,
            LAI_max_tree=LAI_max_tree,
            cb_tree=cb_tree,
            cd_tree=cd_tree,
            ksg_tree=ksg_tree,
            kdd_tree=kdd_tree,
            kws_tree=kws_tree,
            WUE_bare=WUE_bare,
            LAI_max_bare=LAI_max_bare,
            cb_bare=cb_bare,
            cd_bare=cd_bare,
            ksg_bare=ksg_bare,
            kdd_bare=kdd_bare,
            kws_bare=kws_bare,
            **kwds
        )

        for name in self._input_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros("cell", name, units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros("cell", name, units=self._var_units[name])

        self._cell_values = self.grid["cell"]

        self._Blive_ini = self._Blive_init * np.ones(self.grid.number_of_cells)
        self._Bdead_ini = self._Bdead_init * np.ones(self.grid.number_of_cells)

    def initialize(
        self,
        Blive_init=102.0,
        Bdead_init=450.0,
        ETthreshold_up=3.8,
        ETthreshold_down=6.8,
        Tdmax=10.0,
        w=0.55,
        WUE_grass=0.01,
        LAI_max_grass=2.0,
        cb_grass=0.0047,
        cd_grass=0.009,
        ksg_grass=0.012,
        kdd_grass=0.013,
        kws_grass=0.02,
        WUE_shrub=0.0025,
        LAI_max_shrub=2.0,
        cb_shrub=0.004,
        cd_shrub=0.01,
        ksg_shrub=0.002,
        kdd_shrub=0.013,
        kws_shrub=0.02,
        WUE_tree=0.0045,
        LAI_max_tree=4.0,
        cb_tree=0.004,
        cd_tree=0.01,
        ksg_tree=0.002,
        kdd_tree=0.013,
        kws_tree=0.01,
        WUE_bare=0.01,
        LAI_max_bare=0.01,
        cb_bare=0.0047,
        cd_bare=0.009,
        ksg_bare=0.012,
        kdd_bare=0.013,
        kws_bare=0.02,
        **kwds
    ):
        # GRASS = 0; SHRUB = 1; TREE = 2; BARE = 3;
        # SHRUBSEEDLING = 4; TREESEEDLING = 5
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        Blive_init: float, optional
            Initial value for vegetation__live_biomass. Converted to field.
        Bdead_init: float, optional
            Initial value for vegetation__dead_biomass. Coverted to field.
        ETthreshold_up: float, optional
            Potential Evapotranspiration (PET) threshold for
            growing season (mm/d).
        ETthreshold_down: float, optional
            PET threshold for dormant season (mm/d).
        Tdmax: float, optional
            Constant for dead biomass loss adjustment (mm/d).
        w: float, optional
            Conversion factor of CO2 to dry biomass (Kg DM/Kg CO2).
        WUE: float, optional
            Water Use Efficiency - ratio of water used in plant water
            lost by the plant through transpiration (KgCO2Kg-1H2O).
        LAI_max: float, optional
            Maximum leaf area index (m2/m2).
        cb: float, optional
            Specific leaf area for green/live biomass (m2 leaf g-1 DM).
        cd: float, optional
            Specific leaf area for dead biomass (m2 leaf g-1 DM).
        ksg: float, optional
            Senescence coefficient of green/live biomass (d-1).
        kdd: float, optional
            Decay coefficient of aboveground dead biomass (d-1).
        kws: float, optional
            Maximum drought induced foliage loss rate (d-1).
        """
        self._vegtype = self.grid["cell"]["vegetation__plant_functional_type"]
        self._WUE = np.choose(
            self._vegtype,
            [WUE_grass, WUE_shrub, WUE_tree, WUE_bare, WUE_shrub, WUE_tree],
        )
        # Water Use Efficiency  KgCO2kg-1H2O
        self._LAI_max = np.choose(
            self._vegtype,
            [
                LAI_max_grass,
                LAI_max_shrub,
                LAI_max_tree,
                LAI_max_bare,
                LAI_max_shrub,
                LAI_max_tree,
            ],
        )
        # Maximum leaf area index (m2/m2)
        self._cb = np.choose(
            self._vegtype, [cb_grass, cb_shrub, cb_tree, cb_bare, cb_shrub, cb_tree]
        )
        # Specific leaf area for green/live biomass (m2 leaf g-1 DM)
        self._cd = np.choose(
            self._vegtype, [cd_grass, cd_shrub, cd_tree, cd_bare, cd_shrub, cd_tree]
        )
        # Specific leaf area for dead biomass (m2 leaf g-1 DM)
        self._ksg = np.choose(
            self._vegtype,
            [ksg_grass, ksg_shrub, ksg_tree, ksg_bare, ksg_shrub, ksg_tree],
        )
        # Senescence coefficient of green/live biomass (d-1)
        self._kdd = np.choose(
            self._vegtype,
            [kdd_grass, kdd_shrub, kdd_tree, kdd_bare, kdd_shrub, kdd_tree],
        )
        # Decay coefficient of aboveground dead biomass (d-1)
        self._kws = np.choose(
            self._vegtype,
            [kws_grass, kws_shrub, kws_tree, kws_bare, kws_shrub, kws_tree],
        )
        # Maximum drought induced foliage loss rates (d-1)
        self._Blive_init = Blive_init
        self._Bdead_init = Bdead_init
        self._ETthresholdup = ETthreshold_up  # Growth threshold (mm/d)
        self._ETthresholddown = ETthreshold_down  # Dormancy threshold (mm/d)
        self._Tdmax = Tdmax  # Constant for dead biomass loss adjustment
        self._w = w  # Conversion factor of CO2 to dry biomass

        self._Blive_ini = self._Blive_init * np.ones(self.grid.number_of_cells)
        self._Bdead_ini = self._Bdead_init * np.ones(self.grid.number_of_cells)

    def update(self, PETthreshold_switch=0, Tb=24.0, Tr=0.01, **kwds):
        """
        Update fields with current loading conditions.

        Parameters
        ----------
        Tr: float, optional
            Storm duration (hours).
        Tb: float, optional
            Inter-storm duration (hours).
        """
        PETthreshold_ = PETthreshold_switch
        PET = self._cell_values["surface__potential_evapotranspiration_rate"]
        PET30_ = self._cell_values["surface__potential_evapotranspiration_30day_mean"]
        ActualET = self._cell_values["surface__evapotranspiration"]
        Water_stress = self._cell_values["vegetation__water_stress"]

        self._LAIlive = self._cell_values["vegetation__live_leaf_area_index"]
        self._LAIdead = self._cell_values["vegetation__dead_leaf_area_index"]
        self._Blive = self._cell_values["vegetation__live_biomass"]
        self._Bdead = self._cell_values["vegetation__dead_biomass"]
        self._VegCov = self._cell_values["vegetation__cover_fraction"]

        if PETthreshold_ == 1:
            PETthreshold = self._ETthresholdup
        else:
            PETthreshold = self._ETthresholddown

        for cell in range(0, self.grid.number_of_cells):

            WUE = self._WUE[cell]
            LAImax = self._LAI_max[cell]
            cb = self._cb[cell]
            cd = self._cd[cell]
            ksg = self._ksg[cell]
            kdd = self._kdd[cell]
            kws = self._kws[cell]
            # ETdmax = self._ETdmax[cell]
            LAIlive = min(cb * self._Blive_ini[cell], LAImax)
            LAIdead = min(cd * self._Bdead_ini[cell], (LAImax - LAIlive))
            NPP = max((ActualET[cell] / (Tb + Tr)) * WUE * 24.0 * self._w * 1000, 0.001)

            if self._vegtype[cell] == 0:
                if PET30_[cell] > PETthreshold:
                    # Growing Season
                    Bmax = (LAImax - LAIdead) / cb
                    Yconst = 1 / (
                        (1 / Bmax) + (((kws * Water_stress[cell]) + ksg) / NPP)
                    )
                    Blive = (self._Blive_ini[cell] - Yconst) * np.exp(
                        -(NPP / Yconst) * ((Tb + Tr) / 24.0)
                    ) + Yconst
                    Bdead = (
                        self._Bdead_ini[cell]
                        + (Blive - max(Blive * np.exp(-1 * ksg * Tb / 24.0), 0.00001))
                    ) * np.exp(-1 * kdd * min(PET[cell] / self._Tdmax, 1.0) * Tb / 24.0)
                else:  # Senescense
                    Blive = max(
                        self._Blive_ini[cell] * np.exp((-2) * ksg * Tb / 24.0), 1
                    )
                    Bdead = max(
                        (
                            self._Bdead_ini[cell]
                            + (
                                self._Blive_ini[cell]
                                - (
                                    max(
                                        self._Blive_ini[cell]
                                        * np.exp((-2) * ksg * Tb / 24.0),
                                        0.000001,
                                    )
                                )
                            )
                            * np.exp(
                                (-1)
                                * kdd
                                * min(PET[cell] / self._Tdmax, 1.0)
                                * Tb
                                / 24.0
                            ),
                            0.0,
                        )
                    )

            elif self._vegtype[cell] == 3:
                Blive = 0.0
                Bdead = 0.0

            else:
                Bmax = LAImax / cb
                Yconst = 1.0 / (
                    (1.0 / Bmax) + (((kws * Water_stress[cell]) + ksg) / NPP)
                )
                Blive = (self._Blive_ini[cell] - Yconst) * np.exp(
                    -(NPP / Yconst) * ((Tb + Tr) / 24.0)
                ) + Yconst
                Bdead = (
                    self._Bdead_ini[cell]
                    + (Blive - max(Blive * np.exp(-ksg * Tb / 24.0), 0.00001))
                ) * np.exp(-kdd * min(PET[cell] / self._Tdmax, 1.0) * Tb / 24.0)

            LAIlive = min(cb * (Blive + self._Blive_ini[cell]) / 2.0, LAImax)
            LAIdead = min(
                cd * (Bdead + self._Bdead_ini[cell]) / 2.0, (LAImax - LAIlive)
            )
            if self._vegtype[cell] == 0:
                Vt = 1.0 - np.exp(-0.75 * (LAIlive + LAIdead))
            else:
                # Vt = 1 - np.exp(-0.75 * LAIlive)
                Vt = 1.0

            self._LAIlive[cell] = LAIlive
            self._LAIdead[cell] = LAIdead
            self._VegCov[cell] = Vt
            self._Blive[cell] = Blive
            self._Bdead[cell] = Bdead

        self._Blive_ini = self._Blive
        self._Bdead_ini = self._Bdead



# -*- coding: utf-8 -*-
"""
Landlab component to calculate drainage density
"""
from warnings import warn

import numpy as np

from landlab import Component, FieldError

_REQUIRED_FIELDS = (
    "flow__receiver_node",
    "flow__link_to_receiver_node",
    "topographic__steepest_slope",
    "flow__upstream_node_order",
)


class DrainageDensity(Component):

    r"""Calculate drainage density over a DEM.

    Landlab component that implements the distance to channel algorithm of
    Tucker et al., 2001.

    This component requires EITHER a channel__mask array with 1's
    where channels exist and 0's elsewhere, OR a set of coefficients
    and exponents for a slope-area relationship and a
    channelization threshold to compare against that relationship.

    If an array is provided it MUST be of type np.uint8. See the example below
    for how to make such an array.

    The channel__mask array will be assigned to an at-node field with the name
    `channel__mask`. If the channel__mask was originaly created from a passed
    array, a user can update this array to change the mask.

    If the channel__mask is created using an area coefficent,
    slope coefficient, area exponent, slope exponent, and channelization
    threshold, the location of the mask will be re-update when
    calc_drainage_density is called.

    If an area coefficient, :math:`C_A`, a slope coefficent, :math:`C_S`, an
    area exponent, :math:`m_r`, a slope exponent, :math:`n_r`, and
    channelization threshold :math:`T_C` are provided, nodes that meet the
    criteria

    .. math::
        C_A A^{m_r} C_s S^{n_r} > T_c

    where :math:`A` is the drainage density and :math:`S` is the local slope,
    will be marked as channel nodes.

   ``calc_drainage_density`` function returns drainage density for the model
    domain.

    ``calc_drainage_density`` calculates the distance from every node to the
    nearest channel node :math:`L` along the flow line of steepest descent
    (assuming D8 routing).


    This component stores this distance a field, called:
    ``surface_to_channel__minimum_distance``.

    The drainage density is then calculated (after Tucker et al., 2001):

    .. math::

        D_d = \frac{1}{2\overline{L}}

    where :math:`\overline{L}` is the mean L for the model domain.


    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator, FastscapeEroder
    >>> mg = RasterModelGrid((10, 10))
    >>> _ = mg.add_zeros('node', 'topographic__elevation')
    >>> np.random.seed(50)
    >>> noise = np.random.rand(100)
    >>> mg.at_node['topographic__elevation'] += noise
    >>> mg.at_node['topographic__elevation'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0.49460165,  0.2280831 ,  0.25547392,  0.39632991,  0.3773151 ,
        0.99657423,  0.4081972 ,  0.77189399,  0.76053669,  0.31000935,
        0.3465412 ,  0.35176482,  0.14546686,  0.97266468,  0.90917844,
        0.5599571 ,  0.31359075,  0.88820004,  0.67457307,  0.39108745,
        0.50718412,  0.5241035 ,  0.92800093,  0.57137307,  0.66833757,
        0.05225869,  0.3270573 ,  0.05640164,  0.17982769,  0.92593317,
        0.93801522,  0.71409271,  0.73268761,  0.46174768,  0.93132927,
        0.40642024,  0.68320577,  0.64991587,  0.59876518,  0.22203939,
        0.68235717,  0.8780563 ,  0.79671726,  0.43200225,  0.91787822,
        0.78183368,  0.72575028,  0.12485469,  0.91630845,  0.38771099,
        0.29492955,  0.61673141,  0.46784623,  0.25533891,  0.83899589,
        0.1786192 ,  0.22711417,  0.65987645,  0.47911625,  0.07344734,
        0.13896007,  0.11230718,  0.47778497,  0.54029623,  0.95807105,
        0.58379231,  0.52666409,  0.92226269,  0.91925702,  0.25200886,
        0.68263261,  0.96427612,  0.22696165,  0.7160172 ,  0.79776011,
        0.9367512 ,  0.8537225 ,  0.42154581,  0.00543987,  0.03486533,
        0.01390537,  0.58890993,  0.3829931 ,  0.11481895,  0.86445401,
        0.82165703,  0.73749168,  0.84034417,  0.4015291 ,  0.74862   ,
        0.55962945,  0.61323757,  0.29810165,  0.60237917,  0.42567684,
        0.53854438,  0.48672986,  0.49989164,  0.91745948,  0.26287702])
    >>> fr = FlowAccumulator(mg, flow_director='D8')
    >>> fsc = FastscapeEroder(mg, K_sp=.01, m_sp=.5, n_sp=1)
    >>> for x in range(100):
    ...     fr.run_one_step()
    ...     fsc.run_one_step(dt = 10.0)
    ...     mg.at_node['topographic__elevation'][mg.core_nodes] += .01
    >>> channels = np.array(mg.at_node['drainage_area'] > 5, dtype=np.uint8)
    >>> dd = DrainageDensity(mg, channel__mask=channels)
    >>> mean_drainage_density = dd.calc_drainage_density()
    >>> np.isclose(mean_drainage_density, 0.3831100571)
    True

    Alternatively you can pass a set of coefficients to identify the channel
    mask. Next shows the same example as above, but with these coefficients
    provided.

    >>> mg = RasterModelGrid((10, 10))
    >>> _ = mg.add_zeros('node', 'topographic__elevation')
    >>> np.random.seed(50)
    >>> noise = np.random.rand(100)
    >>> mg.at_node['topographic__elevation'] += noise
    >>> fr = FlowAccumulator(mg, flow_director='D8')
    >>> fsc = FastscapeEroder(mg, K_sp=.01, m_sp=.5, n_sp=1)
    >>> for x in range(100):
    ...     fr.run_one_step()
    ...     fsc.run_one_step(dt = 10.0)
    ...     mg.at_node['topographic__elevation'][mg.core_nodes] += .01
    >>> channels = np.array(mg.at_node['drainage_area'] > 5, dtype=np.uint8)
    >>> dd = DrainageDensity(mg,
    ...                      area_coefficient=1.0,
    ...                      slope_coefficient=1.0,
    ...                      area_exponent=1.0,
    ...                      slope_exponent=0.0,
    ...                      channelization_threshold=5)
    >>> mean_drainage_density = dd.calc_drainage_density()
    >>> np.isclose(mean_drainage_density, 0.3831100571)
    True

    """

    _name = "DrainageDensity"

    _input_var_names = (
        "flow__receiver_node",
        "flow__link_to_receiver_node",
        "topographic__steepest_slope",
        "channel__mask",
        "area_coefficient",
        "slope_coefficient",
        "area_exponent",
        "slope_exponent",
        "channelization_threshold",
    )

    _output_var_names = ("surface_to_channel__minimum_distance",)

    _var_units = {
        "flow__receiver_node": "-",
        "flow__link_to_receiver_node": "-",
        "flow__upstream_node_order": "-",
        "topographic__steepest_slope": "-",
        "channel__mask": "-",
        "surface_to_channel__minimum_distance": "m",
    }

    _var_mapping = {
        "flow__receiver_node": "node",
        "flow__link_to_receiver_node": "node",
        "flow__upstream_node_order": "node",
        "topographic__steepest_slope": "node",
        "channel__mask": "node",
        "surface_to_channel__minimum_distance": "node",
    }

    _var_doc = {
        "flow__receiver_node": "Node array of receivers (node that receives flow from current "
        "node)",
        "flow__link_to_receiver_node": "ID of link downstream of each node, which carries the discharge",
        "flow__upstream_node_order": "Node array containing downstream-to-upstream ordered list of "
        "node IDs",
        "topographic__steepest_slope": "Topographic slope at each node",
        "channel__mask": "Logical map of at which grid nodes channels are present",
        "surface_to_channel__minimum_distance": "Distance from each node to the nearest channel",
    }

    def __init__(
        self,
        grid,
        channel__mask=None,
        area_coefficient=None,
        slope_coefficient=None,
        area_exponent=None,
        slope_exponent=None,
        channelization_threshold=None,
        **kwds
    ):
        """Initialize the DrainageDensity component.

        Parameters
        ----------
        grid : ModelGrid
        channel__mask : Array that holds 1's where
            channels exist and 0's elsewhere
        area_coefficient : coefficient to multiply drainage area by,
            for calculating channelization threshold
        slope_coefficient : coefficient to multiply slope by,
            for calculating channelization threshold
        area_exponent : exponent to raise drainage area to,
            for calculating channelization threshold
        slope_exponent : exponent to raise slope to,
            for calculating channelization threshold
        channelization_threshold : threshold value above
            which channels exist
        """
        # Store grid
        self._grid = grid

        for name in _REQUIRED_FIELDS:
            if name not in grid.at_node:
                raise FieldError("{name}: missing required field".format(name=name))

        if grid.at_node["flow__receiver_node"].size != grid.size("node"):
            msg = (
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that DrainageDensity is compatible with "
                "route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )
            raise NotImplementedError(msg)

        if channel__mask is not None:
            if area_coefficient is not None:
                warn(
                    "Channel mask and area "
                    "coefficient supplied. Defaulting "
                    "to channel mask, ignoring area "
                    "coefficient."
                )
            if slope_coefficient is not None:
                warn(
                    "Channel mask and slope "
                    "coefficient supplied. Defaulting "
                    "to channel mask, ignoring slope "
                    "coefficient."
                )
            if area_exponent is not None:
                warn(
                    "Channel mask and area "
                    "exponent supplied. Defaulting "
                    "to channel mask, ignoring area "
                    "exponent."
                )
            if slope_exponent is not None:
                warn(
                    "Channel mask and slope "
                    "exponent supplied. Defaulting "
                    "to channel mask, ignoring slope "
                    "exponent."
                )
            if channelization_threshold is not None:
                warn(
                    "Channel mask and channelization "
                    "threshold supplied. Defaulting "
                    "to channel mask, ignoring "
                    "threshold."
                )
            if grid.number_of_nodes != len(channel__mask):
                raise ValueError(
                    "Length of channel mask is not equal to " "number of grid nodes"
                )

            if "channel__mask" in grid.at_node:
                warn("Existing channel__mask grid field was overwritten.")

            if channel__mask.dtype.type is not np.uint8:
                raise ValueError("mask must by np.uint8")

            self._mask_as_array = True
            self._update_channel_mask = self._update_channel_mask_array
            grid.at_node["channel__mask"] = channel__mask

        if channel__mask is None:
            if area_coefficient is None:
                raise ValueError(
                    "No channel mask and no area "
                    "coefficient supplied. Either "
                    "a channel mask or all 5 threshold "
                    "parameters are needed."
                )
            if slope_coefficient is None:
                raise ValueError(
                    "No channel mask and no slope "
                    "coefficient supplied. Either "
                    "a channel mask or all 5 threshold "
                    "parameters are needed."
                )
            if area_exponent is None:
                raise ValueError(
                    "No channel mask and no area "
                    "exponent supplied. Either "
                    "a channel mask or all 5 threshold "
                    "parameters are needed."
                )
            if slope_exponent is None:
                raise ValueError(
                    "No channel mask and no slope "
                    "exponent supplied. Either "
                    "a channel mask or all 5 threshold "
                    "parameters are needed."
                )
            if channelization_threshold is None:
                raise ValueError(
                    "No channel mask and no channelization "
                    "threshold supplied. Either "
                    "a channel mask or all 5 threshold "
                    "parameters are needed."
                )

            self._mask_as_array = False
            self._update_channel_mask = self._update_channel_mask_values
            self._area_coefficient = area_coefficient
            self._slope_coefficient = slope_coefficient
            self._area_exponent = area_exponent
            self._slope_exponent = slope_exponent
            self._channelization_threshold = channelization_threshold

            self._update_channel_mask()

        # for this component to work with Cython acceleration,
        # the channel_network must be uint8, not bool...
        self._channel_network = grid.at_node["channel__mask"]

        # Flow receivers
        self._flow_receivers = grid.at_node["flow__receiver_node"]

        # Links to receiver nodes
        self._stack_links = grid.at_node["flow__link_to_receiver_node"]

        # Upstream node order
        self._upstream_order = grid.at_node["flow__upstream_node_order"]

        # Distance to channel
        if "surface_to_channel__minimum_distance" in grid.at_node:
            self.distance_to_channel = grid.at_node[
                "surface_to_channel__minimum_distance"
            ]
        else:
            self.distance_to_channel = grid.add_zeros(
                "surface_to_channel__minimum_distance", at="node", dtype=float
            )

    def _update_channel_mask_array(self):
        raise NotImplementedError(
            (
                "If you provided a channel mask to "
                "DrainageDensity, update it by updating the "
                "model grid field channel__mask"
            )
        )

    def _update_channel_mask_values(self):
        channel__mask = (
            self._area_coefficient
            * np.power(self._grid.at_node["drainage_area"], self._area_exponent)
            * self._slope_coefficient
            * np.power(
                self._grid.at_node["topographic__steepest_slope"], self._slope_exponent
            )
        ) > self._channelization_threshold
        self._grid.at_node["channel__mask"] = channel__mask.astype(np.uint8)

    def calc_drainage_density(self):
        """Calculate drainage density. \

        If the channel mask is defined based on slope and area coefficients,
        it will be update based on the current drainage area and slope fields.

        Returns
        -------
        landscape_drainage_density : float (1/m)
            Drainage density over the model domain.
        """
        from .cfuncs import _calc_dists_to_channel

        if self._mask_as_array is False:
            self._update_channel_mask()

        _calc_dists_to_channel(
            self._channel_network,
            self._flow_receivers,
            self._upstream_order,
            self.grid.length_of_d8,
            self._stack_links,
            self.distance_to_channel,
            self.grid.number_of_nodes,
        )
        landscape_drainage_density = 1.0 / (
            2.0
            * np.mean(
                self.grid.at_node["surface_to_channel__minimum_distance"][
                    self.grid.core_nodes
                ]
            )
        )
        # this is THE drainage density
        return landscape_drainage_density



import numpy as np

from landlab.utils.return_array import return_array_at_node

from .cfuncs import calculate_qs_in
from .generalized_erosion_deposition import _GeneralizedErosionDeposition

ROOT2 = np.sqrt(2.0)  # syntactic sugar for precalculated square root of 2
TIME_STEP_FACTOR = 0.5  # factor used in simple subdivision solver
DEFAULT_MINIMUM_TIME_STEP = 0.001  # default minimum time step duration


class ErosionDeposition(_GeneralizedErosionDeposition):
    """
    Erosion-Deposition model in the style of Davy and Lague (2009). It uses a
    mass balance approach across the total sediment mass both in the bed and
    in transport coupled with explicit representation of the sediment
    transport lengthscale (the "xi-q" model) to derive a range of erosional
    and depositional responses in river channels.

    This implementation is close to the Davy & Lague scheme, with a few
    deviations:
    - Sediment porosity is handled explicitly in this implementation.
    - A fraction of the eroded sediment is permitted to enter the wash load,
      and lost to the mass balance (F_f).
    - Here an incision threshold is permitted, where it was not by Davy &
      Lague. It is implemented with an exponentially smoothed form to prevent
      discontinuities in the parameter space. See the
      StreamPowerSmoothThresholdEroder for more documentation.
    - This component uses an "effective" settling velocity, v_s, as one of its
      inputs. This parameter is simply equal to Davy & Lague's `d_star * V`
      dimensionless number.

    Erosion of the bed follows a stream power formulation, i.e.,

    E = K * q ** m_sp * S ** n_sp - optional threshold (see above)

    Note that the transition between transport-limited and detachment-limited
    behavior is controlled by the dimensionless ratio (v_s/r) where r is the
    runoff ratio (Q=Ar). r can be changed in the flow accumulation component
    but is not changed within ErosionDeposition. Because the runoff ratio r
    is not changed within the ErosionDeposition component,  v_s becomes the
    parameter that fundamentally controls response style. Very small v_s will
    lead to a detachment-limited response style, very large v_s will lead to a
    transport-limited response style. v_s == 1 means equal contributions from
    transport and erosion, and a hybrid response as described by Davy & Lague.

    Component written by C. Shobe, K. Barnhart, and G. Tucker.
    """

    _name = "ErosionDeposition"

    _input_var_names = (
        "flow__receiver_node",
        "flow__upstream_node_order",
        "topographic__steepest_slope",
        "surface_water__discharge",
    )

    _output_var_names = "topographic__elevation"

    _var_units = {
        "flow__receiver_node": "-",
        "flow__upstream_node_order": "-",
        "topographic__steepest_slope": "-",
        "surface_water__discharge": "m**2/s",
        "topographic__elevation": "m",
    }

    _var_mapping = {
        "flow__receiver_node": "node",
        "flow__upstream_node_order": "node",
        "topographic__steepest_slope": "node",
        "surface_water__discharge": "node",
        "topographic__elevation": "node",
    }

    _var_doc = {
        "flow__receiver_node": "Node array of receivers (node that receives flow from current "
        "node)",
        "flow__upstream_node_order": "Node array containing downstream-to-upstream ordered list of "
        "node IDs",
        "topographic__steepest_slope": "Topographic slope at each node",
        "surface_water__discharge": "Water discharge at each node",
        "topographic__elevation": "Land surface topographic elevation",
    }

    def __init__(
        self,
        grid,
        K=0.002,
        phi=0.3,
        v_s=1.0,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit=0.0,
        F_f=0.0,
        discharge_field="surface_water__discharge",
        solver="basic",
        dt_min=DEFAULT_MINIMUM_TIME_STEP,
        **kwds
    ):
        """Initialize the ErosionDeposition model.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        K : float, field name, or array
            Erodibility for substrate (units vary).
        phi : float
            Sediment porosity [-].
        v_s : float
            Effective settling velocity for chosen grain size metric [L/T].
        m_sp : float
            Discharge exponent (units vary)
        n_sp : float
            Slope exponent (units vary)
        sp_crit : float, field name, or array
            Critical stream power to erode substrate [E/(TL^2)]
        F_f : float
            Fraction of eroded material that turns into "fines" that do not
            contribute to (coarse) sediment load. Defaults to zero.
        discharge_field : float, field name, or array
            Discharge [L^2/T].
        solver : string
            Solver to use. Options at present include:
                (1) 'basic' (default): explicit forward-time extrapolation.
                    Simple but will become unstable if time step is too large.
                (2) 'adaptive': adaptive time-step solver that estimates a
                    stable step size based on the shortest time to "flattening"
                    among all upstream-downstream node pairs.

        Examples
        ---------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> from landlab.components import DepressionFinderAndRouter
        >>> from landlab.components import ErosionDeposition
        >>> from landlab.components import FastscapeEroder
        >>> np.random.seed(seed = 5000)

        Define grid and initial topography:
            -5x5 grid with baselevel in the lower left corner
            -all other boundary nodes closed
            -Initial topography is plane tilted up to the upper right + noise

        >>> nr = 5
        >>> nc = 5
        >>> dx = 10
        >>> mg = RasterModelGrid((nr, nc), xy_spacing=10.0)
        >>> _ = mg.add_zeros('node', 'topographic__elevation')
        >>> mg['node']['topographic__elevation'] += (mg.node_y/10 +
        ...        mg.node_x/10 + np.random.rand(len(mg.node_y)) / 10)
        >>> mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
        ...                                               left_is_closed=True,
        ...                                               right_is_closed=True,
        ...                                               top_is_closed=True)
        >>> mg.set_watershed_boundary_condition_outlet_id(0,\
                mg['node']['topographic__elevation'], -9999.)
        >>> fsc_dt = 100.
        >>> ed_dt = 1.

        Check initial topography

        >>> mg.at_node['topographic__elevation'] # doctest: +NORMALIZE_WHITESPACE
        array([ 0.02290479,  1.03606698,  2.0727653 ,  3.01126678,  4.06077707,
            1.08157495,  2.09812694,  3.00637448,  4.07999597,  5.00969486,
            2.04008677,  3.06621577,  4.09655859,  5.04809001,  6.02641123,
            3.05874171,  4.00585786,  5.0595697 ,  6.04425233,  7.05334077,
            4.05922478,  5.0409473 ,  6.07035008,  7.0038935 ,  8.01034357])

        Instantiate Fastscape eroder, flow router, and depression finder

        >>> fsc = FastscapeEroder(mg, K_sp=.001, m_sp=.5, n_sp=1)
        >>> fr = FlowAccumulator(mg, flow_director='D8')
        >>> df = DepressionFinderAndRouter(mg)

        Burn in an initial drainage network using the Fastscape eroder:

        >>> for x in range(100):
        ...     fr.run_one_step()
        ...     df.map_depressions()
        ...     flooded = np.where(df.flood_status==3)[0]
        ...     fsc.run_one_step(dt = fsc_dt, flooded_nodes=flooded)
        ...     mg.at_node['topographic__elevation'][0] -= 0.001 #uplift

        Instantiate the E/D component:

        >>> ed = ErosionDeposition(mg, K=0.00001, phi=0.0, v_s=0.001,
        ...                        m_sp=0.5, n_sp = 1.0, sp_crit=0)

        Now run the E/D component for 2000 short timesteps:

        >>> for x in range(2000): #E/D component loop
        ...     fr.run_one_step()
        ...     df.map_depressions()
        ...     flooded = np.where(df.flood_status==3)[0]
        ...     ed.run_one_step(dt = ed_dt, flooded_nodes=flooded)
        ...     mg.at_node['topographic__elevation'][0] -= 2e-4 * ed_dt

        Now we test to see if topography is right:

        >>> np.around(mg.at_node['topographic__elevation'], decimals=3) # doctest: +NORMALIZE_WHITESPACE
        array([-0.477,  1.036,  2.073,  3.011,  4.061,  1.082, -0.08 , -0.065,
           -0.054,  5.01 ,  2.04 , -0.065, -0.065, -0.053,  6.026,  3.059,
           -0.054, -0.053, -0.035,  7.053,  4.059,  5.041,  6.07 ,  7.004,
            8.01 ])
        """
        if grid.at_node["flow__receiver_node"].size != grid.size("node"):
            msg = (
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that ErosionDeposition is compatible with "
                "route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )
            raise NotImplementedError(msg)

        super(ErosionDeposition, self).__init__(
            grid,
            m_sp=m_sp,
            n_sp=n_sp,
            phi=phi,
            F_f=F_f,
            v_s=v_s,
            dt_min=dt_min,
            discharge_field=discharge_field,
        )

        self._grid = grid  # store grid

        # E/D specific inits.

        # K's and critical values can be floats, grid fields, or arrays
        self.K = return_array_at_node(grid, K)
        self.sp_crit = return_array_at_node(grid, sp_crit)

        # Handle option for solver
        if solver == "basic":
            self.run_one_step = self.run_one_step_basic
        elif solver == "adaptive":
            self.run_one_step = self.run_with_adaptive_time_step_solver
            self.time_to_flat = np.zeros(grid.number_of_nodes)
        else:
            raise ValueError(
                "Parameter 'solver' must be one of: " + "'basic', 'adaptive'"
            )

    def _calc_erosion_rates(self):
        """Calculate erosion rates"""
        omega = self.K * self.Q_to_the_m * np.power(self.slope, self.n_sp)
        omega_over_sp_crit = np.divide(
            omega, self.sp_crit, out=np.zeros_like(omega), where=self.sp_crit != 0
        )

        self.erosion_term = omega - self.sp_crit * (1.0 - np.exp(-omega_over_sp_crit))

    def run_one_step_basic(self, dt=1.0, flooded_nodes=[], **kwds):
        """Calculate change in rock and alluvium thickness for
           a time period 'dt'.

        Parameters
        ----------
        dt : float
            Model timestep [T]
        flooded_nodes : array
            Indices of flooded nodes, passed from flow router
        """

        self._calc_hydrology()
        self._calc_erosion_rates()

        self.erosion_term[flooded_nodes] = 0.0
        self.qs_in[:] = 0.0

        # iterate top to bottom through the stack, calculate qs
        # cythonized version of calculating qs_in
        calculate_qs_in(
            np.flipud(self.stack),
            self.flow_receivers,
            self.cell_area_at_node,
            self.q,
            self.qs,
            self.qs_in,
            self.erosion_term,
            self.v_s,
            self.F_f,
        )

        self.depo_rate[:] = 0.0
        self.depo_rate[self.q > 0] = self.qs[self.q > 0] * (
            self.v_s / self.q[self.q > 0]
        )

        # topo elev is old elev + deposition - erosion
        cores = self.grid.core_nodes
        self.topographic__elevation[cores] += (
            (self.depo_rate[cores] - self.erosion_term[cores]) / (1 - self.phi)
        ) * dt

    def run_with_adaptive_time_step_solver(self, dt=1.0, flooded_nodes=[], **kwds):
        """CHILD-like solver that adjusts time steps to prevent slope
        flattening."""

        # Initialize remaining_time, which records how much of the global time
        # step we have yet to use up.
        remaining_time = dt

        z = self._grid.at_node["topographic__elevation"]
        r = self.flow_receivers
        dzdt = np.zeros(len(z))
        cores = self._grid.core_nodes

        first_iteration = True

        # Outer WHILE loop: keep going until time is used up
        while remaining_time > 0.0:

            # Update all the flow-link slopes.
            #
            # For the first iteration, we assume this has already been done
            # outside the component (e.g., by flow router), but we need to do
            # it ourselves on subsequent iterations.
            if not first_iteration:
                # update the link slopes
                self._update_flow_link_slopes()
                # update where nodes are flooded. This shouuldn't happen bc
                # of the dynamic timestepper, but just incase, we update here.
                new_flooded_nodes = np.where(self.slope < 0)[0]
                flooded_nodes = np.asarray(
                    np.unique(np.concatenate((flooded_nodes, new_flooded_nodes))),
                    dtype=np.int64,
                )
            else:
                first_iteration = False

            # Calculate rates of entrainment
            self._calc_hydrology()
            self._calc_erosion_rates()
            self.erosion_term[flooded_nodes] = 0.0
            self.qs_in[:] = 0.0

            # Sweep through nodes from upstream to downstream, calculating Qs.
            calculate_qs_in(
                np.flipud(self.stack),
                self.flow_receivers,
                self.cell_area_at_node,
                self.q,
                self.qs,
                self.qs_in,
                self.erosion_term,
                self.v_s,
                self.F_f,
            )

            # Use Qs to calculate deposition rate at each node.
            self.depo_rate[:] = 0.0
            self.depo_rate[self.q > 0] = self.qs[self.q > 0] * (
                self.v_s / self.q[self.q > 0]
            )

            # Rate of change of elevation at core nodes:
            dzdt[cores] = (self.depo_rate[cores] - self.erosion_term[cores]) / (
                1 - self.phi
            )

            # Difference in elevation between each upstream-downstream pair
            zdif = z - z[r]

            # Rate of change of the *difference* in elevation between each
            # upstream-downstream pair.
            rocdif = dzdt - dzdt[r]

            # (Re)-initialize the array that will contain "time to (almost)
            # flat" for each node (relative to its downstream neighbor).
            self.time_to_flat[:] = remaining_time

            # Find locations where the upstream and downstream node elevations
            # are converging (e.g., the upstream one is eroding faster than its
            # downstream neighbor)
            converging = np.where(rocdif < 0.0)[0]

            # Find the time to (almost) flat by dividing difference by rate of
            # change of difference, and then multiplying by a "safety factor"
            self.time_to_flat[converging] = -(
                TIME_STEP_FACTOR * zdif[converging] / rocdif[converging]
            )

            # Mask out pairs where the source at the same or lower elevation
            # as its downstream neighbor (e.g., because it's a pit or a lake).
            # Here, masking out means simply assigning the remaining time in
            # the global time step.
            self.time_to_flat[np.where(zdif <= 0.0)[0]] = remaining_time
            self.time_to_flat[flooded_nodes] = remaining_time

            # From this, find the maximum stable time step. If it is smaller
            # than our tolerance, report and quit.
            dt_max = max(np.amin(self.time_to_flat), self.dt_min)

            # Finally, apply dzdt to all nodes for a (sub)step of duration
            # dt_max
            z[cores] += dzdt[cores] * dt_max

            # Update remaining time and continue the loop
            remaining_time -= dt_max



#! /bin/env python
import os
import xml.dom.minidom

from .vtkxml import VtkAppendedDataElement

_VALID_ENCODINGS = set(["ascii", "base64", "raw"])
_VALID_FORMATS = set(["ascii", "base64", "raw", "appended"])
_VTK_POSSIBLE_PIECE_SECTIONS = [
    "Points",
    "Coordinates",
    "PointData",
    "Cells",
    "CellData",
]


class VtkError(Exception):
    pass


class InvalidFormatError(VtkError):
    def __init__(self, format):
        self.name = format

    def __str__(self):
        return "%s: Invalid format" % self.name


class InvalidEncodingError(VtkError):
    def __init__(self, encoding):
        self.name = encoding

    def __str__(self):
        return "%s: Invalid encoder" % self.name


def assemble_vtk_document(element):
    root = element["VTKFile"]
    grid = element["Grid"]
    piece = element["Piece"]

    for section in _VTK_POSSIBLE_PIECE_SECTIONS:
        try:
            piece.appendChild(element[section])
        except KeyError:
            pass
    grid.appendChild(piece)
    root.appendChild(grid)

    try:
        root.appendChild(element["AppendedData"])
    except KeyError:
        pass

    return root


def assert_format_is_valid(format_string):
    if format not in _VALID_FORMATS:
        raise InvalidFormatError(format)


def assert_encoding_is_valid(encoding):
    if encoding not in _VALID_ENCODINGS:
        raise InvalidEncodingError(encoding)


class VtkWriter(xml.dom.minidom.Document):
    def __init__(self, **kwds):
        self._format = kwds.pop("format", "ascii")
        self._encoding = kwds.pop("encoding", "ascii")

        assert_format_is_valid(self.format)
        assert_encoding_is_valid(self.encoding)

        if format == "ascii":
            self._encoding = "ascii"

        if self.format == "appended":
            self._data = VtkAppendedDataElement("", encoding=self.encoding)
        else:
            self._data = None

        xml.dom.minidom.Document.__init__(self)

    @property
    def format(self):
        return self._format

    @property
    def encoding(self):
        return self._encoding

    @property
    def data(self):
        return self._data

    def construct_field_elements(self, field):
        raise NotImplementedError()

    def write(self, path, field):
        self.unlink()
        elements = self.construct_field_elements(field)

        if self.data is not None:
            elements["AppendedData"] = self.data

        self.appendChild(assemble_vtk_document(elements))
        self.to_xml(path)

    def to_xml(self, path):
        with open(path, "w") as xml_file:
            xml_file.write(self.toprettyxml())


class VTKDatabase(VtkWriter):
    def write(self, path, **kwargs):
        (base, file) = os.path.split(path)
        (root, ext) = os.path.splitext(file)

        try:
            next_file = "%s_%04d%s" % (root, self._count, ext)
        except NameError:
            self._count = 0
            next_file = "%s_%04d%s" % (root, self._count, ext)

        VtkWriter.write(self, os.path.join(base, next_file), **kwargs)

        self._count += 1


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from .eventlayers import EventLayers, EventLayersMixIn
from .materiallayers import MaterialLayers, MaterialLayersMixIn

__all__ = ["EventLayers", "EventLayersMixIn", "MaterialLayers", "MaterialLayersMixIn"]



#!/usr/bin/env python
"""
.. codeauthor:: Eric Hutton <huttone@colorado.edu>

.. sectionauthor:: Eric Hutton <huttone@colorado.edu>
"""

from .dupuit_percolator import GroundwaterDupuitPercolator

__all__ = ["GroundwaterDupuitPercolator"]



from collections import OrderedDict

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal
from six.moves import StringIO

from landlab import HexModelGrid, RadialModelGrid, RasterModelGrid
from landlab.grid.create import (
    _parse_args_kwargs,
    add_boundary_conditions,
    add_field_from_function,
    as_list_of_tuples,
    create_grid,
    grid_from_dict,
    norm_grid_description,
)

SIMPLE_PARAMS_STR = """
grid:
  RasterModelGrid:
    args: [4, 5]
    xy_spacing: [3, 4]
    fields:
      node:
        topographic__elevation:
          plane:
            - point: [1, 1, 1]
              normal: [-2, -1, 1]
    boundary_conditions:
      - set_closed_boundaries_at_grid_edges:
        - True
        - True
        - True
        - True
"""


def test_no_grid_value():
    dict_like = {"foo": "bar"}
    with pytest.raises(ValueError):
        create_grid(dict_like)


def test_bad_grid_name():
    dict_like = {"grid": "MagicModelGrid"}
    with pytest.raises(ValueError):
        create_grid(dict_like, section="grid")


def test_two_grid_types_as_dict():
    dict_like = {
        "grid": OrderedDict([("RasterModelGrid", [(4, 5)]), ("HexModelGrid", [6, 7])])
    }

    # dict_like = {"grid": {"RasterModelGrid": [(4, 5)], "HexModelGrid": [6, 7]}}
    # with pytest.raises(ValueError):
    grids = create_grid(dict_like, section="grid")
    assert len(grids) == 2
    assert isinstance(grids[0], RasterModelGrid)
    assert isinstance(grids[1], HexModelGrid)


def test_two_grid_types_as_list():
    dict_like = {"grid": [{"RasterModelGrid": [(4, 5)]}, {"HexModelGrid": [6, 7]}]}

    grids = create_grid(dict_like, section="grid")
    assert len(grids) == 2
    assert isinstance(grids[0], RasterModelGrid)
    assert isinstance(grids[1], HexModelGrid)


def test_bad_field_location():
    dict_like = {
        "grid": {
            "RasterModelGrid": {
                "args": [(4, 5)],
                "fields": {"at_bad_location": {"foo": "bar"}},
            }
        }
    }
    with pytest.raises(ValueError):
        create_grid(dict_like, section="grid")


def test_bad_field_function():
    dict_like = {
        "grid": {
            "RasterModelGrid": [
                (4, 5),
                {
                    "fields": {
                        "node": {"new_field_name": {"not_a_function": ["bar", "spam"]}}
                    }
                },
            ]
        }
    }
    with pytest.raises(ValueError):
        create_grid(dict_like, section="grid")


def test_simple_create(tmpdir):
    """Load parameters from YAML-formatted file."""
    with tmpdir.as_cwd():
        with open("params.yaml", "w") as fp:
            fp.write(SIMPLE_PARAMS_STR)
        with pytest.deprecated_call():
            mg = create_grid("./params.yaml", section="grid")

    assert mg.number_of_nodes == 20
    assert "topographic__elevation" in mg.at_node

    x_of_node = np.array(
        [
            0.0,
            3.0,
            6.0,
            9.0,
            12.0,
            0.0,
            3.0,
            6.0,
            9.0,
            12.0,
            0.0,
            3.0,
            6.0,
            9.0,
            12.0,
            0.0,
            3.0,
            6.0,
            9.0,
            12.0,
        ]
    )

    status_at_node = np.array(
        [4, 4, 4, 4, 4, 4, 0, 0, 0, 4, 4, 0, 0, 0, 4, 4, 4, 4, 4, 4], dtype=np.uint8
    )
    topographic__elevation = np.array(
        [
            [-2.0, 4.0, 10.0, 16.0, 22.0],
            [2.0, 8.0, 14.0, 20.0, 26.0],
            [6.0, 12.0, 18.0, 24.0, 30.0],
            [10.0, 16.0, 22.0, 28.0, 34.0],
        ]
    )

    assert_array_equal(mg.x_of_node, x_of_node)
    assert_array_equal(status_at_node, mg.status_at_node)
    assert_array_equal(
        topographic__elevation,
        np.round(mg.at_node["topographic__elevation"].reshape(mg.shape), decimals=2),
    )


def test_esri_ascii_create(datadir):
    filename = str(datadir / "4_x_3_no_nodata_value.asc")
    dict_like = {
        "grid": {
            "RasterModelGrid": [
                (4, 3),
                {
                    "xy_spacing": 10,
                    "xy_of_lower_left": (1, 2),
                    "fields": {
                        "node": {
                            "topographic__elevation": {"read_esri_ascii": [filename]}
                        }
                    },
                },
            ]
        }
    }
    mg = create_grid(dict_like, section="grid")

    x_of_node = np.array(
        [1.0, 11.0, 21.0, 1.0, 11.0, 21.0, 1.0, 11.0, 21.0, 1.0, 11.0, 21.0]
    )
    y_of_node = np.array(
        [2.0, 2.0, 2.0, 12.0, 12.0, 12.0, 22.0, 22.0, 22.0, 32.0, 32.0, 32.0]
    )
    status_at_node = np.array([1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1], dtype=np.uint8)
    topographic__elevation = np.array(
        [9.0, 10.0, 11.0, 6.0, 7.0, 8.0, 3.0, 4.0, 5.0, 0.0, 1.0, 2.0]
    )

    assert_array_equal(mg.x_of_node, x_of_node)
    assert_array_equal(mg.y_of_node, y_of_node)
    assert_array_equal(status_at_node, mg.status_at_node)
    assert_array_equal(topographic__elevation, mg.at_node["topographic__elevation"])


def test_read_netcdf_create(datadir):
    filename = str(datadir / "test-netcdf4.nc")
    dict_like = {
        "grid": {
            "RasterModelGrid": [
                (4, 3),
                {
                    "fields": {
                        "node": {"surface__elevation": {"read_netcdf": [filename]}}
                    }
                },
            ]
        }
    }
    mg = create_grid(dict_like, section="grid")
    x_of_node = np.array([0.0, 1.0, 2.0, 0.0, 1.0, 2.0, 0.0, 1.0, 2.0, 0.0, 1.0, 2.0])
    y_of_node = np.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])
    status_at_node = np.array([1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1], dtype=np.uint8)
    surface__elevation = np.array(
        [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0]
    )

    assert_array_equal(mg.x_of_node, x_of_node)
    assert_array_equal(mg.y_of_node, y_of_node)
    assert_array_equal(status_at_node, mg.status_at_node)
    assert_array_equal(surface__elevation, mg.at_node["surface__elevation"])


def test_two_boundary_condition_dicts():
    dict_like = {
        "grid": {"RasterModelGrid": [(4, 3)]},
        "boundary_conditions": [
            {
                "set_closed_boundaries_at_grid_edges": [True, True, True, True],
                "not_a_function": [True, False],
            }
        ],
    }
    with pytest.raises(ValueError):
        create_grid(dict_like)


def test_bad_boundary_condition_functions(datadir):
    filename = str(datadir / "bad_boundary.yaml")
    with pytest.raises(ValueError):
        create_grid(filename, section="grid")


def test_parse_args_kwds():
    assert _parse_args_kwargs(["foo", dict(bar="baz")]) == (("foo",), {"bar": "baz"})


def test_parse_args_kwds_without_kwds():
    assert _parse_args_kwargs(["foo"]) == (("foo",), {})


def test_parse_args_kwds_without_args():
    assert _parse_args_kwargs([dict(bar="baz")]) == ((), {"bar": "baz"})


def test_parse_args_kwds_with_multiple_args():
    assert _parse_args_kwargs(["foo", "bar", dict(bar="baz")]) == (
        ("foo", "bar"),
        {"bar": "baz"},
    )


def test_parse_args_kwds_as_dict():
    assert _parse_args_kwargs({"args": "foo", "bar": "baz"}) == _parse_args_kwargs(
        ["foo", dict(bar="baz")]
    )
    assert _parse_args_kwargs({"args": ["foo"], "bar": "baz"}) == _parse_args_kwargs(
        {"args": "foo", "bar": "baz"}
    )
    assert _parse_args_kwargs({"args": "foo"}) == (("foo",), {})
    assert _parse_args_kwargs({"bar": "baz"}) == ((), {"bar": "baz"})


@pytest.mark.parametrize("xy_of_lower_left", ((-10.0, 100.0), None))
@pytest.mark.parametrize("xy_spacing", (10.0, (100.0, 200.0), None))
def test_grid_from_dict_raster(xy_of_lower_left, xy_spacing):
    kwds = {}
    if xy_of_lower_left is not None:
        kwds["xy_of_lower_left"] = xy_of_lower_left
    if xy_spacing is not None:
        kwds["xy_spacing"] = xy_spacing

    expected = RasterModelGrid((3, 4), **kwds)
    actual = grid_from_dict("RasterModelGrid", [(3, 4), kwds])

    assert_array_almost_equal(expected.x_of_node, actual.x_of_node)
    assert_array_almost_equal(expected.y_of_node, actual.y_of_node)


@pytest.mark.parametrize("orientation", ("horizontal", "vertical", None))
@pytest.mark.parametrize("node_layout", ("hex", "rect", None))
@pytest.mark.parametrize("xy_of_lower_left", ((-10.0, 100.0), None))
def test_grid_from_dict_hex(orientation, node_layout, xy_of_lower_left):
    kwds = dict(shape=(3, 4))
    if orientation is not None:
        kwds["orientation"] = orientation
    if node_layout is not None:
        kwds["node_layout"] = node_layout
    if xy_of_lower_left is not None:
        kwds["xy_of_lower_left"] = xy_of_lower_left

    expected = HexModelGrid(**kwds)
    actual = grid_from_dict("HexModelGrid", kwds)

    assert_array_almost_equal(expected.x_of_node, actual.x_of_node)
    assert_array_almost_equal(expected.y_of_node, actual.y_of_node)


@pytest.mark.parametrize("xy_of_center", ((-10.0, 100.0), None))
def test_grid_from_dict_radial(xy_of_center):
    kwds = {}
    if xy_of_center is not None:
        kwds["xy_of_center"] = xy_of_center
    expected = RadialModelGrid(3, 0.5, **kwds)
    actual = grid_from_dict("RadialModelGrid", (3, 0.5, kwds))

    assert_array_almost_equal(expected.x_of_node, actual.x_of_node)
    assert_array_almost_equal(expected.y_of_node, actual.y_of_node)


@pytest.mark.parametrize(
    "to_list",
    (
        lambda l: ((k, v) for k, v in l),
        lambda l: tuple((k, v) for k, v in l),
        lambda l: [[k, v] for k, v in l],
        lambda l: [{k: v} for k, v in l],
        OrderedDict,
    ),
)
def test_as_list_of_tuples(to_list):
    expected = [
        ("RasterModelGrid", "raster_data"),
        ("hex_model_grid", "hex_data"),
        ("radial_model_grid", "radial_data"),
    ]
    grid_list = to_list(expected)
    assert as_list_of_tuples(grid_list) == expected


def test_as_list_of_tuples_just_one():
    assert as_list_of_tuples(("foo", "bar")) == [("foo", "bar")]
    assert as_list_of_tuples(["foo", "bar"]) == [("foo", "bar")]
    assert as_list_of_tuples({"foo": "bar"}) == [("foo", "bar")]


def test_as_list_of_tuples_mixed_types():
    expected = [
        ("RasterModelGrid", "raster_data"),
        ("hex_model_grid", "hex_data"),
        ("radial_model_grid", "radial_data"),
    ]
    grid_list = [
        ("RasterModelGrid", "raster_data"),
        OrderedDict(
            [("hex_model_grid", "hex_data"), ("radial_model_grid", "radial_data")]
        ),
    ]
    assert as_list_of_tuples(grid_list) == expected


def test_create_grid():
    contents = StringIO(
        """
RasterModelGrid:
  args: [3, 4]
  xy_spacing: 2.0
  xy_of_lower_left: [1.0, 2.0]
  fields:
    node:
      elevation:
        constant:
          value: 1.0
"""
    )
    expected = RasterModelGrid((3, 4), xy_spacing=2.0, xy_of_lower_left=(1, 2))
    with pytest.deprecated_call():
        actual = create_grid(contents)
    assert_array_almost_equal(expected.x_of_node, actual.x_of_node)
    assert_array_almost_equal(expected.y_of_node, actual.y_of_node)


def test_create_grid_multiple():
    contents = StringIO(
        """
grids:
    - RasterModelGrid:
      - [3, 4]
      - xy_spacing: 2.0
        xy_of_lower_left: [1.0, 2.0]
    - RasterModelGrid:
        args: [3, 4]
        xy_spacing: 2.0
        xy_of_lower_left: [1.0, 2.0]
    - - RasterModelGrid
      - args: [3, 4]
        xy_spacing: 2.0
        xy_of_lower_left: [1.0, 2.0]
"""
    )
    expected = RasterModelGrid((3, 4), xy_spacing=2.0, xy_of_lower_left=(1, 2))
    with pytest.deprecated_call():
        grids = create_grid(contents, section="grids")
    assert len(grids) == 3
    for actual in grids:
        assert_array_almost_equal(expected.x_of_node, actual.x_of_node)
        assert_array_almost_equal(expected.y_of_node, actual.y_of_node)


@pytest.mark.parametrize("to_list", (list, tuple))
@pytest.mark.parametrize("at", ("node", "link", "patch", "corner", "face", "cell"))
def test_add_field_from_function_constant(at, to_list):
    grid = RasterModelGrid((20, 30))
    function = to_list(("constant", dict(value=1.0)))
    add_field_from_function(grid, "z", function, at=at)
    assert_array_almost_equal(grid[at]["z"], 1.0)


@pytest.mark.parametrize("at", ("node", "link", "patch", "corner", "face", "cell"))
def test_add_field_from_function_constant_as_dict(at):
    grid = RasterModelGrid((20, 30))
    add_field_from_function(grid, "z", {"constant": dict(value=10.0)}, at=at)
    assert_array_almost_equal(grid[at]["z"], 10.0)


@pytest.mark.parametrize(
    "to_list",
    (
        lambda l: ((k, v) for k, v in l),
        lambda l: tuple((k, v) for k, v in l),
        lambda l: [[k, v] for k, v in l],
        lambda l: [{k: v} for k, v in l],
    ),
)
@pytest.mark.parametrize("at", ("node", "link", "patch", "corner", "face", "cell"))
def test_add_field_from_function_with_list(at, to_list):
    grid = RasterModelGrid((20, 30))
    functions = to_list(
        [
            ("constant", dict(value=1.0)),
            ("constant", dict(value=2.0)),
            ("constant", dict(value=3.0)),
            ("constant", dict(value=4.0)),
            ("constant", dict(value=5.0)),
        ]
    )
    add_field_from_function(grid, "z", functions, at=at)
    assert_array_almost_equal(grid[at]["z"], 15.0)


@pytest.mark.parametrize(
    "to_list",
    (
        lambda l: ((k, v) for k, v in l),
        lambda l: tuple((k, v) for k, v in l),
        lambda l: [[k, v] for k, v in l],
        lambda l: [{k: v} for k, v in l],
    ),
)
def test_add_boundary_conditions(to_list):
    boundary_conditions = to_list(
        [("set_closed_boundaries_at_grid_edges", (True, True, True, True))]
    )
    grid = RasterModelGrid((3, 4))
    add_boundary_conditions(grid, boundary_conditions=boundary_conditions)
    assert_array_equal(
        grid.status_at_node,
        [
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CORE,
            grid.BC_NODE_IS_CORE,
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CLOSED,
            grid.BC_NODE_IS_CLOSED,
        ],
    )


def test_norm_grid_description():
    expected = {"args": ["arg0", "arg1"]}
    assert norm_grid_description(["arg0", "arg1"]) == expected
    assert norm_grid_description(expected) == expected



#! /usr/bin/env python
import os

import numpy as np

from landlab import RasterModelGrid
from landlab.io import read_esri_ascii
from landlab.io.netcdf import read_netcdf


def test_save_esri_ascii(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=2.0)
    grid.add_field("node", "air__temperature", np.arange(20.0))

    with tmpdir.as_cwd():
        grid.save("test.asc", format="esri-ascii")
        assert os.path.isfile("test.asc")


def test_add_extension(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=2.0)
    grid.add_field("node", "air__temperature", np.arange(20.0))

    with tmpdir.as_cwd():
        grid.save("test", format="esri-ascii")
        assert os.path.isfile("test.asc")

    with tmpdir.as_cwd():
        grid.save("test", format="netcdf")
        assert os.path.isfile("test.nc")


def test_replace_extension(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=2.0)
    grid.add_field("node", "air__temperature", np.arange(20.0))

    with tmpdir.as_cwd():
        grid.save("test.nc", format="esri-ascii")
        assert os.path.isfile("test.asc")

    with tmpdir.as_cwd():
        grid.save("test.asc", format="netcdf")
        assert os.path.isfile("test.nc")


def test_guess_format(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=2.0)
    grid.add_field("node", "air__temperature", np.arange(20.0))

    with tmpdir.as_cwd():
        grid.save("test.asc")
        assert os.path.isfile("test.asc")
        read_esri_ascii("test.asc")

    with tmpdir.as_cwd():
        grid.save("test.nc")
        assert os.path.isfile("test.nc")
        read_netcdf("test.nc")


def test_names_keyword_as_str(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=2.0)
    grid.add_field("air__temperature", np.arange(20.0), at="node")
    grid.add_field("land_surface__elevation", np.arange(20.0), at="node")

    with tmpdir.as_cwd():
        grid.save("test.asc", names="land_surface__elevation")
        assert os.path.isfile("test.asc")
        read_esri_ascii("test.asc")


def test_names_keyword_as_list(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=2.0)
    grid.add_field("air__temperature", np.arange(20.0), at="node")
    grid.add_field("land_surface__elevation", np.arange(20.0), at="node")

    with tmpdir.as_cwd():
        grid.save("test.asc", names=["land_surface__elevation", "air__temperature"])
        files = ["test_land_surface__elevation.asc", "test_air__temperature.asc"]
        for fname in files:
            assert os.path.isfile(fname)
            read_esri_ascii(fname)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 11 10:13:38 2017

@author: margauxmouchene
"""


import numpy as np

from landlab import Component


class TransportLengthHillslopeDiffuser(Component):

    """
    Hillslope diffusion component in the style of Carretier et al. (2016,
        ESurf), and Davy and Lague (2009)

    dz/dt = - E + D (+ Uplift)
    D = qs / L
    E = k * S
    L = dx / (1 - (S / Sc)^2)

    Works on regular raster-type grid (RasterModelGrid, dx=dy).
    To be coupled with FlowDirectorSteepest for the calculation of steepest
     slope at each timestep.

    Component written by Margaux Mouchene, 2017

    Parameters
    ----------
    grid : ModelGrid
        Landlab ModelGrid object
    erodibility: float
        Erodibility coefficient [L/T]
    slope_crit: float (default=1.)
        Critical slope [L/L]

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowDirectorSteepest
    >>> from landlab.components import TransportLengthHillslopeDiffuser

    Define grid and initial topography:
        - 3x5 grid
        - east and west boundaries are open, north and south are closed
        - Initial topography is plane at base level on the boundaries and
            1m of elevation elsewhere (core)

    >>> mg = RasterModelGrid((5, 5))
    >>> mg.set_closed_boundaries_at_grid_edges(False, True, False, True)
    >>> z = np.array([0., 0., 0., 0., 0.,
    ...               0., 1., 1., 1., 0.,
    ...               0., 1., 1., 1., 0.,
    ...               0., 1., 1., 1., 0.,
    ...               0., 0., 0., 0., 0.])
    >>> _ = mg.add_field('node', 'topographic__elevation', z)

    Instantiate Flow director (steepest slope type) and TL hillslope diffuser

    >>> fdir = FlowDirectorSteepest(mg)
    >>> tl_diff = TransportLengthHillslopeDiffuser(mg, \
                                                   erodibility=0.001,\
                                                   slope_crit=0.6)

    Run the components for ten short timepsteps

    >>> for t in range(10):
    ...     fdir.run_one_step()
    ...     tl_diff.run_one_step(1.)

    Check final topography

    >>> np.allclose(
    ...     mg.at_node['topographic__elevation'],
    ...     np.array([ 0.,  0.        ,  0.        ,  0.        ,  0.,
    ...                0.,  0.96175283,  0.99982519,  0.96175283,  0.,
    ...                0.,  0.96175283,  0.99982519,  0.96175283,  0.,
    ...                0.,  0.96175283,  0.99982519,  0.96175283,  0.,
    ...                0.,  0.        ,  0.        ,  0.        ,  0.]))
    True
    """

    _name = "TransportLengthHillslopeDiffuser"

    _input_var_names = set(
        ("topographic__elevation", "flow__receiver_node", "topographic__steepest_slope")
    )

    _output_var_names = set(
        (
            "topographic__elevation",
            "sediment__deposition_rate",
            "sediment__transfer_rate",
            "sediment__deposition_coeff",
            "sediment__flux_in",
            "sediment__flux_out",
            "sediment__erosion_rate",
        )
    )

    _var_units = {
        "topographic__elevation": "m",
        "flow__receiver_node": "-",
        "topographic__steepest_slope": "m/m",
        "sediment__deposition_rate": "m/yr",
        "sediment__transfer_rate": "m/yr",
        "sediment__deposition_coeff": "-",
        "sediment__flux_in": "m/yr",
        "sediment__flux_out": "m/yr",
        "sediment__erosion_rate": "m/yr",
    }

    _var_mapping = {
        "topographic__elevation": "node",
        "flow__receiver_node": "node",
        "topographic__steepest_slope": "node",
        "sediment__deposition_rate": "node",
        "sediment__transfer_rate": "node",
        "sediment__deposition_coeff": "node",
        "sediment__flux_in": "node",
        "sediment__flux_out": "node",
        "sediment__erosion_rate": "node",
    }

    _var_doc = {
        "topographic__elevation": "Elevation of the ground surface",
        "flow__receiver_node": "Node array of receivers (node that receives flow from "
        "current node)",
        "topographic__steepest_slope": "Steepest gradient of the ground surface at each node",
        "sediment__deposition_rate": "Deposition rate on node",
        "sediment__transfer_rate": "Rate of transferred sediment across a node (incoming "
        "sediment - deposited sediment on node)",
        "sediment__deposition_coeff": "Fraction of incoming sediment that is deposited on the node",
        "sediment__flux_in": "Incoming sediment rate on node (=qs/dx)",
        "sediment__flux_out": "Outgoing sediment rate on node = sediment eroded on"
        " node + sediment transported across node from upstream",
        "sediment__erosion_rate": "Erosion rate on node",
    }

    def __init__(self, grid, erodibility, slope_crit=1.0):

        """Initialize Diffuser.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        erodibility: float
            Erodibility coefficient [L/T]
        slope_crit: float (default=1.)
            Critical slope [L/L]
        """
        if grid.at_node["flow__receiver_node"].size != grid.size("node"):
            msg = (
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that TransportLengthHillslopeDiffuser is compatible "
                "with route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )
            raise NotImplementedError(msg)

        # Store grid and parameters
        self._grid = grid
        self.k = erodibility
        self.slope_crit = slope_crit

        # Create fields:
        # Elevation
        if "topographic__elevation" in self.grid.at_node:
            self.elev = self.grid.at_node["topographic__elevation"]
        else:
            self.elev = self.grid.add_zeros("node", "topographic__elevation")

        # Deposition
        if "sediment__deposition_rate" in self.grid.at_node:
            self.depo = self.grid.at_node["sediment__deposition_rate"]
        else:
            self.depo = self.grid.add_zeros("node", "sediment__deposition_rate")

        # Transferred sediments (crossing over node)
        if "sediment__transfer_rate" in self.grid.at_node:
            self.trans = self.grid.at_node["sediment__transfer_rate"]
        else:
            self.trans = self.grid.add_zeros("node", "sediment__transfer_rate")

        # Transport coefficient
        if "sediment__deposition_coeff" in self.grid.at_node:
            self.d_coeff = self.grid.at_node["sediment__deposition_coeff"]
        else:
            self.d_coeff = self.grid.add_zeros("node", "sediment__deposition_coeff")

        # Flux in
        if "sediment__flux_in" in self.grid.at_node:
            self.flux_in = self.grid.at_node["sediment__flux_in"]
        else:
            self.flux_in = self.grid.add_zeros("node", "sediment__flux_in")

        # Flux out
        if "sediment__flux_out" in self.grid.at_node:
            self.flux_out = self.grid.at_node["sediment__flux_out"]
        else:
            self.flux_out = self.grid.add_zeros("node", "sediment__flux_out")

        # Erosion
        if "sediment__erosion_rate" in self.grid.at_node:
            self.erosion = self.grid.at_node["sediment__erosion_rate"]
        else:
            self.erosion = self.grid.add_zeros("node", "sediment__erosion_rate")

    def tldiffusion(self, dt):
        """Calculate hillslope diffusion for a time period 'dt'.

        Parameters:
        grid : ModelGrid
            Landlab ModelGrid object
        dt: float (time)
            The imposed timestep.
        """

        # Reset erosion, depo, trans and flux_in to 0
        self.erosion[:] = 0.0
        self.depo[:] = 0.0
        self.trans[:] = 0.0
        self.flux_in[:] = 0.0

        # Downstream steepest slope at node:
        self.steepest = self.grid.at_node["topographic__steepest_slope"]
        # On each node, node ID of downstream receiver node
        # (on node (i), ID of node that receives flow from node (i)):
        self.receiver = self.grid.at_node["flow__receiver_node"]

        dx = self.grid.dx
        cores = self.grid.core_nodes

        # Calculate influx rate on node i  = outflux of nodes
        # whose receiver is i
        for i in self.grid.core_nodes:
            self.flux_in[self.receiver[i]] += self.flux_out[i]

            # Calculate transport coefficient
            # When S ~ Scrit, d_coeff is set to "infinity", for stability and
            # so that there is no deposition
            if self.steepest[i] >= self.slope_crit:
                self.d_coeff[i] = 1000000000.0
            else:
                self.d_coeff[i] = 1 / (
                    1 - (np.power(((self.steepest[i]) / self.slope_crit), 2))
                )

        # Calculate deposition rate on node
        self.depo[cores] = self.flux_in[cores] / self.d_coeff[cores]

        # Calculate erosion rate on node (positive value)
        # If S > Scrit, erosion is simply set for the slope to return to Scrit
        # Otherwise, erosion is slope times erodibility coefficent
        for i in self.grid.core_nodes:
            if self.steepest[i] > self.slope_crit:
                self.erosion[i] = dx * (self.steepest[i] - self.slope_crit) / (100 * dt)
            else:
                self.erosion[i] = self.k * self.steepest[i]

            # Update elevation
            self.elev[i] += (-self.erosion[i] + self.depo[i]) * dt

        # Calculate transfer rate over node
        self.trans[cores] = self.flux_in[cores] - self.depo[cores]

        # Calculate outflux rate
        self.flux_out[:] = self.erosion + self.trans

    def run_one_step(self, dt):
        """
        Advance transport length-model hillslope diffusion component
        by one time step of size dt and tests for timestep stability.

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.
        """
        self.tldiffusion(dt)

        # Test code stability for timestep dt
        # Raise unstability error if local slope is reversed by erosion
        # and deposition during a timestep dt
        elev_dif = self.elev - self.elev[self.receiver]
        s = elev_dif[np.where(self.grid.at_node["flow__sink_flag"] == 0)]
        if np.any(s < -1) is True:
            raise ValueError(
                "The component is unstable" " for such a large timestep " "on this grid"
            )
        else:
            pass
